from dli.client.collection_functions import CollectionFunctions
from dli.client.dataset_mixin import DatasetMixin
from dli.client.package_functions import PackageFunctions


"""
Aggregate mixin of catalogue entities
"""
class EntityMixin(CollectionFunctions, PackageFunctions, DatasetMixin):
    pass


import logging
import os
from unittest import skip

import six
from backports import tempfile
from mock import call, patch

from dli.client import utils
from dli.client.exceptions import (CatalogueEntityNotFoundException,
                                   DownloadFailed,
                                   InsufficientPrivilegesException,
                                   InvalidPayloadException,
                                   MissingMandatoryArgumentException)

from .common import SdkIntegrationTestCase, build_fake_s3fs

logger = logging.getLogger(__name__)


class DatafileFunctionsTestCase(SdkIntegrationTestCase):

    def setUp(self):
        super(DatafileFunctionsTestCase, self).setUp()

        self.package_id = self.create_package("test_datafile_functions")
        self.dataset_name = "test_datafile_functions"
        self.builder = self.dataset_builder(
            self.package_id, self.dataset_name)

    def create_s3_dataset(self, bucket_name, prefix="prefix"):
        return self.register_s3_dataset(self.package_id, self.dataset_name, bucket_name, prefix=prefix)

    def create_dummy_datafile(self):
        dataset = self.register_s3_dataset(self.package_id, self.dataset_name, "test-bucket", "prefix")
        datafile = self.client.register_s3_datafile(
            dataset.dataset_id,
            "dummy",
            [os.path.relpath(__file__)],
            "prefix",
            data_as_of='2000-01-01'
        )
        return datafile

    def test_get_datafile_with_none_id_throws(self):
        with self.assertRaises(MissingMandatoryArgumentException):
            self.client.get_datafile(None)

    def test_get_unknown_datafile_raises_exception(self):
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.get_datafile("unknown")

    def test_get_unknown_s3_datafile_returns_error(self):
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.get_s3_datafile("unknown")

    def test_can_get_s3_datafile(self):
        s3_dataset = self.create_s3_dataset("test-bucket")
        datafile = self.client.register_s3_datafile(
            s3_dataset.dataset_id,
            "test_get_s3_datafile",
            [os.path.relpath(__file__)],
            "prefix",
            data_as_of='2000-01-01'
        )

        s3_datafile = self.client.get_s3_datafile(datafile.datafile_id)
        self.assertEqual(s3_datafile.datafile_id, datafile.datafile_id)

    def test_register_datafile_metadata(self):
        files = [{'path': "/path/to/file/A", 'size': 99999},
                 {'path': "/path/to/file/B", 'size': 88888}]
        dataset = self.client.register_dataset(
            self.builder.with_external_storage(
                location="jdbc://connectionstring:1232/my-db"
            )
        )
        datafile = self.client.register_datafile_metadata(
            dataset.dataset_id,
            "test_register_dataset_metadata",
            files,
            data_as_of='2000-01-01'
        )

        self.assertEqual(datafile.dataset_id, dataset.dataset_id)
        self.assertEqual(datafile.files, files)

    def test_register_datafile_metadata_fails_when_data_as_of_is_invalid(self):
        files = [{'path': "/path/to/file/A", 'size': 99999},
                 {'path': "/path/to/file/B", 'size': 88888}]
        dataset = self.client.register_dataset(
            self.builder.with_external_storage(
                location="jdbc://connectionstring:1232/my-db"
            )
        )

        with self.assertRaises(ValueError):
            self.client.register_datafile_metadata(
                dataset.dataset_id,
                "test_register_dataset_metadata",
                files,
                data_as_of='2000-111-01'
            )

    def test_register_datafile_metadata_fails_when_no_files_provided(self):
        dataset = self.client.register_dataset(self.builder.with_external_storage(
            location="jdbc://connectionstring:1232/my-db"))
        with self.assertRaises(Exception):
            self.client.register_datafile_metadata(
                dataset.dataset_id,
                "test_register_datafile_metadata_fails_when_no_files_provided",
                files=[]
            )

    def test_update_datafile_fails_for_unknown_datafile(self):
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.edit_datafile_metadata("unknown")

    def test_register_s3_datafile_can_create_datafile_uploading_files(self):
        dataset = self.create_s3_dataset("dev-ihsm-dl-pkg-test")
        file = os.path.relpath(__file__)  # upload ourselves as a dataset
        datafile = self.client.register_s3_datafile(
            dataset.dataset_id,
            "test_register_s3_datafile_can_create_datafile_uploading_files",
            [file],
            "prefix",
            data_as_of='2000-01-01'
        )

        self.assertIsNotNone(datafile)
        self.assertEqual(datafile.dataset_id, dataset.dataset_id)
        self.assertEqual(datafile.files, [
            {"path": "s3://dev-ihsm-dl-pkg-test/prefix/" +
                os.path.basename(file)}
        ])

    def test_register_datafile_metadata_on_datalake_dataset(self):
        dataset = self.create_s3_dataset("dev-ihsm-dl-pkg-test")
        file = os.path.relpath(__file__)  # upload ourselves as a dataset
        datafile = self.client.register_datafile_metadata(
            dataset.dataset_id,
            "test_register_datafile_metadata_on_datalake_dataset",
            [
                {"path": "s3://dev-ihsm-dl-pkg-test/prefix/test"}
            ],
            data_as_of='2000-01-01'
        )

        self.assertIsNotNone(datafile)
        self.assertEqual(datafile.dataset_id, dataset.dataset_id)

    def test_update_datafile_metadata_fails_when_data_as_of_is_invalid(self):
        datafile = self.create_dummy_datafile()
        with self.assertRaises(ValueError):
            self.client.edit_datafile_metadata(
                datafile.datafile_id,
                name="correct name",
                data_as_of="dataAsOf"
            )

    def test_update_datafile_merges_changes_with_existing_datafile(self):
        datafile = self.create_dummy_datafile()
        updated = self.client.edit_datafile_metadata(
            datafile.datafile_id,
            name="correct name"
        )

        self.assertEqual(datafile.datafile_id, updated.datafile_id)
        self.assertEqual(datafile.dataset_id, updated.dataset_id)
        self.assertEqual(datafile.files, updated.files)
        self.assertEqual(updated.name, "correct name")

    def test_can_delete_datafile(self):
        datafile = self.create_dummy_datafile()
        # delete the datafile
        self.client.delete_datafile(datafile.datafile_id)
        # can't read it back
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.get_datafile(datafile.datafile_id)

    def test_delete_unknown_datafile_raises_exception(self):
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.delete_datafile("unknown")

    def test_can_add_files_to_existing_datafile(self):
        file = os.path.relpath(__file__)  # upload ourselves as a dataset
        file2 = '../test_sandbox/samples/data/AAPL.csv'

        dataset = self.create_s3_dataset("dev-ihsm-dl-pkg-test")
        datafile = self.client.register_s3_datafile(
            dataset.dataset_id,
            "test_can_add_files_to_existing_datafile",
            [file],
            "prefix",
            data_as_of='2000-01-01'
        )

        updated = self.client.add_files_to_datafile(
            datafile.datafile_id, 'prefix', [file2])

        # countEqual asserts that the collections match disrespect of order
        # talk about naming stuff
        six.assertCountEqual(
            self,
            [
                {"path": "s3://dev-ihsm-dl-pkg-test/prefix/" +
                    os.path.basename(file)},
                {"path": "s3://dev-ihsm-dl-pkg-test/prefix/" +
                    os.path.basename(file2)}
            ],
            updated.files
        )

    def test_add_files_to_unknown_datafile_raises_exception(self):
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.add_files_to_datafile(
                'unknown', 'prefix', ["/path/to/file/A"])

    def test_register_datafile_to_nonexisting_dataset_raises_exception(self):
        with self.assertRaises(CatalogueEntityNotFoundException) as context:
            self.client.register_datafile_metadata(
                "unkknown",
                "test_register_datafile_to_nonexisting_dataset_raises_exception",
                [{'path': "/path/to/file/A"}, {'path': "/path/to/file/B"}],
                data_as_of='2000-01-01'
            )
            self.assertTrue(
                'Dataset with id unknown not found' in context.exception)


class DownloadDatafileTestCase(SdkIntegrationTestCase):

    def setUp(self):
        super(DownloadDatafileTestCase, self).setUp()
        # create a package
        self.package_id = self.create_package("test download datafile")
        self.dataset_name = "test download datafile"
        self.builder = self.dataset_builder(
            self.package_id, self.dataset_name)

    def create_s3_dataset(self, bucket_name, prefix="prefix"):
        return self.register_s3_dataset(self.package_id, self.dataset_name, bucket_name, prefix)
    
    def test_download_datafile_for_unknown_datafile_fails(self):
        with self.assertRaises(Exception):
            self.client.download_datafile("unknown")

    def test_download_datafile_retrieves_all_files_in_datafile(self):
        dataset = self.create_s3_dataset("dev-ihsm-dl-pkg-test")
        datafile = self.client.register_s3_datafile(
            dataset.dataset_id,
            "test_download_dataset_retrieves_all_files_in_dataset",
            [
                '../test_sandbox/samples/data/AAPL.csv',
                '../test_sandbox/samples/data/MSFT.csv'
            ],
            "prefix",
            '2000-01-01'
        )

        with tempfile.TemporaryDirectory() as dest:
            self.client.download_datafile(datafile.datafile_id, dest)

            # validate we got the expected calls
            self.s3_download_mock.assert_has_calls([
                call("s3://dev-ihsm-dl-pkg-test/prefix/MSFT.csv", dest),
                call("s3://dev-ihsm-dl-pkg-test/prefix/AAPL.csv", dest),
            ],
                any_order=True
            )

    def test_download_dataset_keeps_going_if_a_file_in_the_dataset_fails(self):
        def _download(_, file, dest):
            if file.endswith("AAPL.csv"):
                raise Exception("")

        dataset = self.create_s3_dataset("dev-ihsm-dl-pkg-test")
        datafile = self.client.register_s3_datafile(
            dataset.dataset_id,
            "test_download_dataset_keeps_going_if_a_file_in_the_dataset_fails",
            [
                '../test_sandbox/samples/data/AAPL.csv',
                '../test_sandbox/samples/data/MSFT.csv'
            ],
            "prefix",
            '2000-01-01'
        )

        with self.assertRaises(DownloadFailed):
            with patch('dli.client.s3.Client.download_files_from_s3_path', _download) as s3_download:
                with tempfile.TemporaryDirectory() as dest:
                    self.client.download_datafile(datafile.datafile_id, dest)

                    # validate we got the expected calls
                    s3_download.assert_has_calls([
                        call("s3://dev-ihsm-dl-pkg-test/prefix/MSFT.csv", dest),
                        call("s3://dev-ihsm-dl-pkg-test/prefix/AAPL.csv", dest),
                    ],
                        any_order=True
                    )


@patch("dli.client.s3.build_s3fs", build_fake_s3fs)
class RegisterDatafileTestCase(SdkIntegrationTestCase):

    def set_s3_client_mock(self):
        pass

    def create_and_get_dataset_id(self):
        package_id = self.create_package(
            "RegisterDatafileTestCase"
        )
        dataset = self.register_s3_dataset(
            package_id, "test", "dev-ihsm-dl-pkg-test"
        )

        return dataset.dataset_id

    def test_can_upload_datafile_when_provided_folder_with_relative_path(self):
        sample_data = os.path.join(
            os.path.dirname(__file__),
            'resources/yahoo'
        )

        datafile = self.client.register_s3_datafile(
            self.create_and_get_dataset_id(),
            "test_can_upload_datafile_when_provided_folder_with_relative_path",
            [sample_data],
            "prefix",
            '2000-01-01'
        )

        # assert the files were uploaded and that
        # their sizes have been resolved
        self.assertIn({
            "path": "s3://dev-ihsm-dl-pkg-test/prefix/AAPL.csv",
            "size": os.path.getsize(os.path.join(sample_data, "AAPL.csv"))
        },
            datafile.files
        )
        self.assertIn({
            "path": "s3://dev-ihsm-dl-pkg-test/prefix/MSFT.csv",
            "size": os.path.getsize(os.path.join(sample_data, "MSFT.csv"))
        },
            datafile.files
        )

    def test_can_upload_datafile_when_provided_s3_prefix_without_trailing_slash(self):
        sample_data = os.path.join(
            os.path.dirname(__file__),
            'resources/yahoo/AAPL.csv'
        )

        # Prefix Without trailing slash
        datafile = self.client.register_s3_datafile(
            self.create_and_get_dataset_id(),
            "test_can_upload_datafile_when_provided_s3_prefix_with_or_without_trailing_slashes 1",
            [sample_data],
            "prefix",
            '2000-01-01'
        )

        self.assertIn({
            "path": "s3://dev-ihsm-dl-pkg-test/prefix/AAPL.csv",
            "size": os.path.getsize(sample_data)
        },
            datafile.files
        )

    def test_can_upload_datafile_with_data_as_of_date_not_in_iso_format(self):
        sample_data = os.path.join(
            os.path.dirname(__file__),
            'resources/yahoo/AAPL.csv'
        )

        datafile = self.client.register_s3_datafile(
            self.create_and_get_dataset_id(),
            "test_can_upload_datafile_with_data_as_of_date_not_in_iso_format 1",
            [sample_data],
            "prefix",
            '3210-1-1'
        )

        self.assertEquals(self.client.get_datafile(datafile.datafile_id).data_as_of, '3210-01-01')
        self.assertIn({
            "path": "s3://dev-ihsm-dl-pkg-test/prefix/AAPL.csv",
            "size": os.path.getsize(sample_data)
        },
            datafile.files
        )

    def test_cannot_upload_datafile_with_invalid_data_as_of(self):
        sample_data = os.path.join(
            os.path.dirname(__file__),
            'resources/yahoo/AAPL.csv'
        )

        with self.assertRaises(ValueError):
            self.client.register_s3_datafile(
                self.create_and_get_dataset_id(),
                "test_can_upload_datafile_with_data_as_of_date_not_in_iso_format 1",
                [sample_data],
                "prefix",
                '2019-11-123'
        )


    def test_can_upload_datafile_files_recursively(self):
        sample_data = os.path.join(
            os.path.dirname(__file__),
            'resources/stocks'
        )

        datafile = self.client.register_s3_datafile(
            self.create_and_get_dataset_id(),
            "test_can_upload_datafile_files_recursively",
            [sample_data],
            "prefix",
            '2000-01-01'
        )

        # assert the files were uploaded and that
        # their sizes have been resolved
        self.assertIn({
            "path": "s3://dev-ihsm-dl-pkg-test/prefix/readme.txt",
            "size": os.path.getsize(os.path.join(sample_data, "readme.txt"))
        },
            datafile.files
        )
        self.assertIn({
            "path": "s3://dev-ihsm-dl-pkg-test/prefix/microsoft/MSFT.csv",
            "size": os.path.getsize(os.path.join(sample_data, "microsoft/MSFT.csv"))
        },
            datafile.files
        )
        self.assertIn({
            "path": "s3://dev-ihsm-dl-pkg-test/prefix/microsoft/AAPL.csv",
            "size": os.path.getsize(os.path.join(sample_data, "microsoft/AAPL.csv"))
        },
            datafile.files
        )
        self.assertIn({
            "path": "s3://dev-ihsm-dl-pkg-test/prefix/microsoft/cortana/master.txt",
            "size": os.path.getsize(os.path.join(sample_data, "microsoft/cortana/master.txt"))
        },
            datafile.files
        )

    def test_download_datafile_should_keep_folder_structure_as_in_datafile(self):
        sample_data = os.path.join(
            os.path.dirname(__file__),
            'resources/stocks'
        )

        datafile = self.client.register_s3_datafile(
            self.create_and_get_dataset_id(),
            "test_download_datafile_should_keep_folder_structure_as_in_datafile",
            [sample_data],
            "prefix",
            '2000-01-01'
        )

        with tempfile.TemporaryDirectory() as dest:
            self.client.download_datafile(datafile.datafile_id, dest)

            self.assertTrue(utils.path_for(dest, "prefix").exists())
            self.assertTrue(utils.path_for(
                dest, "prefix", "microsoft").exists())
            self.assertTrue(utils.path_for(
                dest, "prefix", "microsoft").exists())
            self.assertTrue(utils.path_for(
                dest, "prefix", "microsoft", "AAPL.csv").exists())
            self.assertTrue(utils.path_for(
                dest, "prefix", "microsoft", "MSFT.csv").exists())
            self.assertTrue(utils.path_for(
                dest, "prefix", "microsoft", "cortana").exists())
            self.assertTrue(utils.path_for(
                dest, "prefix", "microsoft", "cortana", "master.txt").exists())



import logging
import os
import tempfile
from unittest import TestCase

import six
from backports import tempfile as tf

from dli.client.exceptions import S3FileDoesNotExist, DownloadDestinationNotValid
from dli.client.s3 import Client, S3DatafileWrapper
from tests import localstack_helper
from tests.common import build_fake_s3fs

logger = logging.getLogger(__name__)


class S3ClientTestCase(TestCase):

    def setUp(self):
        self.target = Client("key", "secret", "token")
        self.target.s3fs = build_fake_s3fs("key", "sectet", "token")
        s3client = localstack_helper.get_s3_client()
        s3client.create_bucket(Bucket="bucket")

        def cleanup():
            files_to_delete = self.target.s3fs.walk("bucket")
            s3client.delete_objects(Bucket="bucket", Delete={"Objects": [{"Key": f.replace("bucket/", "", 1)} for f in files_to_delete]})
            s3client.delete_bucket(Bucket="bucket")

        self.addCleanup(cleanup)

    def test_download_file_validates_path_exists_in_s3(self):
        with self.assertRaises(S3FileDoesNotExist):
            with tf.TemporaryDirectory() as dest:
                self.target.download_files_from_s3_path("s3://unknown/file", dest)

    def test_download_file_validates_destination_is_a_directory(self):
        self.target.s3fs.put(__file__, "s3://bucket/file")
        with self.assertRaises(DownloadDestinationNotValid):
            with tempfile.NamedTemporaryFile() as cm:
                self.target.download_files_from_s3_path("s3://bucket/file", cm.name)

    def test_download_file_creates_destination_directory_if_it_doesnt_exist(self):
        self.target.s3fs.put(
            __file__,
            os.path.join("s3://bucket/location/", os.path.basename(__file__))
        )

        with tf.TemporaryDirectory() as dest:
            dest = os.path.join(dest, "dir1", "dir2")
            self.target.download_files_from_s3_path(
                "s3://bucket/location/test_s3_client.py",
                dest
            )

            self.assertTrue(os.path.exists(dest) and os.path.isdir(dest))
            self.assertTrue(os.path.exists(os.path.join(dest, "location", "test_s3_client.py")))

    def test_download_file_can_download_folders(self):
        self.target.s3fs.put(__file__, "s3://bucket/tdfcdf/file1.txt")
        self.target.s3fs.put(__file__, "s3://bucket/tdfcdf/subdir/file2.txt")
        self.target.s3fs.put(__file__, "s3://bucket/tdfcdf/subdir/subdir/file3.txt")

        with tf.TemporaryDirectory() as dest:
            self.target.download_files_from_s3_path("s3://bucket/tdfcdf", dest)

            self.assertTrue(os.path.exists(dest))
            self.assertTrue(os.path.exists(os.path.join(dest, "tdfcdf", "file1.txt")))
            self.assertTrue(os.path.exists(os.path.join(dest, "tdfcdf", "subdir", "file2.txt")))
            self.assertTrue(os.path.exists(os.path.join(dest, "tdfcdf", "subdir", "subdir", "file3.txt")))

    def test_download_file_can_download_single_file(self):
        self.target.s3fs.put(__file__, "s3://bucket/tdfcdf/file1.txt")

        with tf.TemporaryDirectory() as dest:
            self.target.download_files_from_s3_path("s3://bucket/tdfcdf/file1.txt", dest)

            self.assertTrue(os.path.exists(dest))
            self.assertTrue(os.path.exists(os.path.join(dest, "tdfcdf", "file1.txt")))


class S3DatafileWrapperTestCase(TestCase):

    def setUp(self):
        self.s3fs = build_fake_s3fs("key", "sectet", "token")

    def test_files_returns_files_in_datafile(self):
        files = [
            "s3://bucket/tfrfid/a",
            "s3://bucket/tfrfid/b",
            "s3://bucket/tfrfid/c",
        ]

        for f in files:
            self.s3fs.touch(f)

        datafile = {
            "files": [{"path": f} for f in files]
        }

        target = S3DatafileWrapper(datafile, self.s3fs)
        six.assertCountEqual(
            self,
            target.files,
            [f.replace("s3://", "") for f in files]
        )

    def test_files_will_recurse_directories(self):
        files = [
            "s3://bucket/tfwrd/a/aa",
            "s3://bucket/tfwrd/b/bb/bbb",
            "s3://bucket/tfwrd/c/cc/ccc/cccc1",
            "s3://bucket/tfwrd/c/cc/ccc/cccc2",
        ]

        # create sample files
        for f in files:
            self.s3fs.touch(f)

        datafile = {
            "files": [
                {'path': 's3://bucket/tfwrd/a'},
                {'path': 's3://bucket/tfwrd/b'},
                {'path': 's3://bucket/tfwrd/c'}
            ]
        }

        target = S3DatafileWrapper(datafile, self.s3fs)
        six.assertCountEqual(
            self,
            target.files,
            [f.replace("s3://", "") for f in files]
        )

    def test_only_files_in_datafile_are_displayed(self):
        files = [
            "s3://bucket/tofidad/a",
            "s3://bucket/tofidad/b",
            "s3://bucket/tofidad/c",
            "s3://bucket/tofidad/d"
        ]

        # create sample files
        for f in files:
            self.s3fs.touch(f)

        datafile = {
            "files": [{"path": f} for f in files[0:1]]
        }
        target = S3DatafileWrapper(datafile, self.s3fs)
        six.assertCountEqual(
            self,
            target.files,
            [f.replace("s3://", "") for f in files[0:1]]
        )

    def test_can_open_file(self):
        files = [
            "s3://bucket/tcof/a"
        ]
        # create sample files
        for f in files:
            self.s3fs.touch(f)
            with self.s3fs.open(f, mode="wb") as s3f:
                s3f.write(b"test 1")
                s3f.flush()

        datafile = {
            "files":  [{"path": f} for f in files]
        }

        target = S3DatafileWrapper(datafile, self.s3fs)

        with target.open_file("bucket/tcof/a") as s3file:
            self.assertIsNotNone(s3file)
            self.assertEquals(s3file.read(), b"test 1")

    def test_unknown_file_is_handled_gracefully(self):
        files = [
            "s3://bucket/tufihg/a"
        ]
        # create sample files
        for f in files:
            self.s3fs.touch(f)

        datafile = {
            "files": [{"path": f} for f in files]
        }

        target = S3DatafileWrapper(datafile, self.s3fs)

        with self.assertRaises(S3FileDoesNotExist):
            target.open_file("bucket/unknown/file")



import unittest
import requests
from requests import Session
import json
import httpretty
from dli.client.dli_request_factory_factory import DliRequestFactoryFactory
from dli import __version__
from dli.client.exceptions import (
    DatalakeException,
    InsufficientPrivilegesException,
    UnAuthorisedAccessException
)


class DliRequestFactoryFactoryTestCase(unittest.TestCase):

    @httpretty.activate
    def test_response_403_raises_InsufficientPrivilegesException(self):
        response_text = 'Insufficient Privileges'
        httpretty.register_uri(httpretty.GET, 'http://dummy.com/test', status=403, body=response_text)

        dli_request_factory_factory = DliRequestFactoryFactory('http://dummy.com')

        with self.assertRaises(InsufficientPrivilegesException):
            response = Session().send(dli_request_factory_factory.request_factory(method='GET', url='/test').prepare())

    @httpretty.activate
    def test_response_401_raises_UnAuthorisedAccessException(self):
        response_text = 'UnAuthorised Access'
        httpretty.register_uri(httpretty.GET, 'http://dummy.com/test', status=401, body=response_text)

        dli_request_factory_factory = DliRequestFactoryFactory('http://dummy.com')

        with self.assertRaises(UnAuthorisedAccessException):
            response = Session().send(dli_request_factory_factory.request_factory(method='GET', url='/test').prepare())

    @httpretty.activate
    def test_response_500_raises_DatalakeException(self):
        response_text = 'Datalake server error'
        httpretty.register_uri(httpretty.GET, 'http://dummy.com/test', status=500, body=response_text)

        dli_request_factory_factory = DliRequestFactoryFactory('http://dummy.com')

        with self.assertRaises(DatalakeException):
            response = Session().send(dli_request_factory_factory.request_factory(method='GET', url='/test').prepare())

    @httpretty.activate
    def test_sdk_version_is_included_in_header(self):
        httpretty.register_uri(httpretty.GET, 'http://dummy.com/test', status=200, body="response")
        dli_request_factory_factory = DliRequestFactoryFactory('http://dummy.com')

        # issue a request
        Session().send(dli_request_factory_factory.request_factory(method='GET', url='/test').prepare())

        request = httpretty.last_request()
        self.assertTrue("X-Data-Lake-SDK-Version" in request.headers)
        self.assertEqual(request.headers["X-Data-Lake-SDK-Version"], str(__version__))



import logging

from dli.client.exceptions import (
    CatalogueEntityNotFoundException,
    MissingMandatoryArgumentException,
)

from dli.siren import siren_to_entity, siren_to_dict


logger = logging.getLogger(__name__)


class AutoRegMetadataFunctions:

    @property
    def __root(self):
        return self.ctx.memoized(
            self.get_root_siren().auto_reg_metadatas_root,
            "auto_reg_metadatas_root"
        )

    def set_auto_registration_metadata(
        self,
        dataset_id,
        path_template,
        name_template,
        as_of_date_template,
        active,
        sns_topic_for_s3_events,
        handle_files=False,
        has_auto_registry_failure_notifications=True
    ):
        """
        Submit a request to set up the auto registration metadata for a dataset.

        See description for each parameter, and whether they are optional or mandatory.

        :param str dataset_id: Dataset ID for which the auto registration metadata is being set up.
        :param str path_template: Path template for the files stored under the dataset.
        :param str name_template: Name template for the datafiles registered under the dataset.
        :param str as_of_date_template: As of date template for the datafiles registered under the dataset.
        :param bool active: Flag to indicate the auto registration status of the dataset
                            i.e. True => Active, False => Inactive or disabled
        :param str sns_topic_for_s3_events: Name of SNS topic where the S3 notification events are published when
                            files are added, updated or deleted from the S3 bucket for this dataset.
        :param bool handle_files: Optional. Flag to indicate whether the individual files (for e.g. parquet part files
                                    in case of a parquet dataset) are to be registered under the datafile for the dataset. Defaults to false.
        :param bool has_auto_registry_failure_notifications: Optional. Controls whether auto-registry sends email notifications on error. Defaults to true.
        :returns: Created auto registration metadata object.
        :rtype: collections.namedtuple

        - **Sample**

        .. code-block:: python

                # For example if the files for dataset id `xyz` are stored in the s3 bucket along
                `region=US/as_of_date=2019-02-18`

                auto_reg_metadata = client.set_auto_registration_metadata(
                    dataset_id="xyz",
                    path_template="region={{ region }}/as_of_date={{ year }}-{{ month }}-{{ day }}",
                    name_template="Datafile_name_{{ region }}_{{ year }}-{{ month }}-{{ day }}",
                    as_of_date_template="{{ year }}-{{ month }}-{{ day }}",
                    active=True,
                    sns_topic_for_s3_events='s3-notifications-sns-topic'
                )
        """
        
        if not dataset_id:
            raise MissingMandatoryArgumentException('dataset_id')

        dataset = self.get_dataset(id=dataset_id)

        fields = {
            'datasetId': dataset.dataset_id,
            'pathTemplate': path_template,
            'nameTemplate': name_template,
            'asOfDateTemplate': as_of_date_template,
            'active': active,
            'snsTopicForS3Events': sns_topic_for_s3_events,
            'handlePartFiles': handle_files,
            "hasAutoRegistryFailureNotifications": has_auto_registry_failure_notifications
        }

        payload = {k: v for k, v in fields.items() if v is not None}
        return siren_to_entity(self.__root.create_auto_reg_metadata(__json=payload))

    def get_auto_registration_metadata(self, dataset_id=None, auto_reg_metadata_id=None):
        """
        Fetches the auto registration metadata for a dataset.

        :param str dataset_id: Optional. The dataset id for which auto registration metadata is being fetched.
                            Either this or auto_reg_metadata_id is required for the look up
        :param str auto_reg_metadata_id: Optional. The id of the auto registration metadata.
                            Either this or dataset_id is required for the look up
        :returns: The auto registration metadata object.
        :rtype: collections.namedtuple

        - **Sample**

        .. code-block:: python

                # Look up by dataset id
                auto_reg_metadata = client.get_auto_registration_metadata('my_dataset_id')
                # or equivalent
                auto_reg_metadata = client.auto_reg_metadata(dataset_id='my_dataset_id')


                # If auto reg metadata id is known
                auto_reg_metadata = client.auto_reg_metadata(auto_reg_metadata_id='my_auto_reg_id')
        """

        if dataset_id and auto_reg_metadata_id:
            return siren_to_entity(
                self._get_auto_reg_metadata(
                    auto_reg_metadata_id=auto_reg_metadata_id,
                    dataset_id=dataset_id
                )
            )
        elif auto_reg_metadata_id:
            return siren_to_entity(
                self._get_auto_reg_metadata(
                    auto_reg_metadata_id=auto_reg_metadata_id
                )
            )
        elif dataset_id:
            return siren_to_entity(
                self._get_auto_reg_metadata(
                    dataset_id=dataset_id
                )
            )

        raise ValueError("Either dataset id or auto reg metadata id must be specified to look up auto registration details for the dataset")

    def _get_auto_reg_metadata(self, **kwargs):

        auto_reg_metadata = self.__root.get_auto_reg_metadata(**kwargs)

        if not auto_reg_metadata:
            raise CatalogueEntityNotFoundException('AutoRegMetadata', params=kwargs)

        return auto_reg_metadata

    def edit_auto_registration_metadata(
        self,
        auto_reg_metadata_id,
        path_template=None,
        name_template=None,
        as_of_date_template=None,
        active=None,
        sns_topic_for_s3_events=None,
        handle_files=None,
        has_auto_registry_failure_notifications=None
    ):
        """
        Edits existing auto registration metadata for a dataset.
        Fields passed as ``None`` will retain their original value.

        :param str auto_reg_metadata_id: The id of the auto registration metadata we want to modify.
        :param str path_template: Optional. Path template for the files stored under the dataset.
        :param str name_template: Optional. Name template for the datafiles registered under the dataset.
        :param str as_of_date_template: Optional. As of date template for the datafiles registered under the dataset.
        :param bool active: Optional. Boolean flag to indicate the auto registration status of the dataset
                            i.e. True => Active, False => Inactive or disabled
        :param str sns_topic_for_s3_events: Optional. Name of SNS topic where the S3 notification events are published when
                            files are added, updated or deleted from the S3 bucket for this dataset.
        :param bool handle_files: Optional. Flag to indicate whether the individual files (for e.g. parquet part files
                                    in case of a parquet dataset) are to be registered under the datafile for the dataset.
        :param bool has_auto_registry_failure_notifications: Optional. Controls whether auto-registry sends email notifications on error.
        :returns: Updated auto registration metadata object
        :rtype: collections.namedtuple

        - **Sample**

        .. code-block:: python

                # For example edit metadata to make it inactive

                updated_auto_reg_metadata = client.edit_auto_registration_metadata(
                    auto_reg_metadata_id="test-auto-reg-metadata-id",
                    active=False
                )
        """

        if not auto_reg_metadata_id:
            raise MissingMandatoryArgumentException('auto_reg_metadata_id')

        auto_reg_metadata = self._get_auto_reg_metadata(auto_reg_metadata_id=auto_reg_metadata_id)

        fields = {
            'datasetId': auto_reg_metadata.datasetId,
            'pathTemplate': path_template,
            'nameTemplate': name_template,
            'asOfDateTemplate': as_of_date_template,
            'active': active,
            'snsTopicForS3Events': sns_topic_for_s3_events,
            'handlePartFiles': handle_files,
            "hasAutoRegistryFailureNotifications": has_auto_registry_failure_notifications
        }

        # clean up any unknown fields, and update the entity
        auto_reg_metadata_dict = siren_to_dict(auto_reg_metadata)
        for key in list(auto_reg_metadata_dict.keys()):
            if key not in fields:
                del auto_reg_metadata_dict[key]

        payload = {k: v for k, v in fields.items() if v is not None}
        auto_reg_metadata_dict.update(payload)

        # perform the update and return the resulting entity
        return siren_to_entity(auto_reg_metadata.edit_auto_reg_metadata(__json=auto_reg_metadata_dict))

    def delete_auto_registration_metadata(self, auto_reg_metadata_id):
        """
        Marks auto registration metadata as deleted.

        :param str auto_reg_metadata_id: the id for the metadata we want to delete.

        :returns: None
        - **Sample**

        .. code-block:: python

                client.delete_auto_registration_metadata(auto_reg_metadata_id)
        """

        if not auto_reg_metadata_id:
            raise MissingMandatoryArgumentException('auto_reg_metadata_id')

        auto_reg_metadata = self._get_auto_reg_metadata(auto_reg_metadata_id=auto_reg_metadata_id)
        if not auto_reg_metadata:
            raise CatalogueEntityNotFoundException('AutoRegMetadata', params=auto_reg_metadata_id)

        auto_reg_metadata.delete_auto_reg_metadata(auto_reg_metadata_id=auto_reg_metadata_id)


from dli.siren import siren_to_entity
from dli.client.utils import ensure_count_is_valid


class SearchFunctions:
    """
    Catalogue search functions.
    """

    @property
    def __root(self):
        return self.ctx.memoized(self.get_root_siren)

    def _search_my_entities(self, entity, search_func, term, count):
        ensure_count_is_valid(count)

        # replicating UI behavior, for empty term we want an empty search
        if term is None or term == "":
            return []

        result = search_func(query=term, page_size=count)
        return result.get_entities(rel=entity)

    def search(self, term, count=100):
        """
        Search across all catalogue entities given a particular set of keywords.

        :param str term: The search term.
        :param int count: Optional. The amount of results to be returned.

        :returns: A list of Catalogue entities
        :rtype: list[collections.namedtuple]

        - **Sample**

        .. code-block:: python

                automotive_catalogue_entities = client.search(
                    term="Automotive",
                    count=100
                )

        """
        results = self._search_my_entities("", self.__root.search, term, count)
        return [siren_to_entity(e) for e in results]

    def search_packages(self, term, count=100):
        """
        Search across packages in the catalogue given a particular set of keywords.

        :param str term: The search term.
        :param int count: Optional. The amount of results to be returned. Defaults to 100.

        :returns: A list of packages
        :rtype: list[collections.namedtuple]

        - **Sample**

        .. code-block:: python

                automotive_packages = client.search_packages(
                    term="Automotive",
                    count=100
                )

        """
        packages = self._search_my_entities("package", self.__root.search_packages, term, count)
        return [siren_to_entity(p) for p in packages]

    def search_collections(self, term, count=100):
        """
        Search across collections in the catalogue given a particular set of keywords.

        :param str term: The search term.
        :param int count: Optional. The amount of results to be returned. Defaults to 100.

        :returns: A list of collections
        :rtype: list[collections.namedtuple]

        - **Sample**

        .. code-block:: python

                automotive_collections = client.search_collections(
                    term="Automotive",
                    count=100
                )

        """
        collections = self._search_my_entities("collection", self.__root.search_collections, term, count)
        return [siren_to_entity(c) for c in collections]

    def search_datasets(self, term, count=100):
        """
        Search across datasets in the catalogue given a particular set of keywords.

        :param str term: The search term.
        :param int count: Optional. The amount of results to be returned. Defaults to 100.

        :returns: A list of datasets
        :rtype: list[dataset.namedtuple]

        - **Sample**

        .. code-block:: python

                results = client.search_datasets(
                    term="CDS",
                    count=100
                )

        """
        datasets = self._search_my_entities("dataset", self.__root.search_datasets, term, count)
        return [siren_to_entity(c) for c in datasets]



from dli.client.dataset_functions import DatasetFunctions
from dli.client.datafile_functions import DatafileFunctions
from dli.client.auto_reg_metadata_functions import AutoRegMetadataFunctions


"""
Aggregate mixin of Dataset and related entities.
"""

class DatasetMixin(DatasetFunctions, DatafileFunctions, AutoRegMetadataFunctions):
    pass


import requests
from requests_toolbelt.adapters import host_header_ssl

from dli.client.client_mixin import ClientMixin
from dli.client.context import Context

class AuthenticationFailure(Exception):
    """
    An exception wrapping an authentication failure response. If the response
    had a payload, that payload is reported as the exception message, otherwise
    a generic error message is returned.
    """
    GENERIC_ERROR_MESSAGE = (
        'API key is not valid or has expired. Please generate a new key in the Catalogue UI and try again.'
    )

    def __init__(self, response=None, message=None):
        self.response = response
        self.message = message

    def __str__(self):
        if self.response and self.response.text:
            return self.response.text
        elif self.message:
            return self.message

        return AuthenticationFailure.GENERIC_ERROR_MESSAGE


class DliClient(ClientMixin):
    """
    Definition of a client. This client mixes in utility functions for
    manipulating collections, packages, datasets and datafiles.
    """
    def __init__(self, api_key, api_root, host=None):
        self.api_key = api_key
        self.api_root = api_root
        self.host = host
        self._ctx = self._init_ctx()

    def _init_ctx(self):
        auth_key = _get_auth_key(
            self.api_key,
            self.api_root,
            self.host
        )

        return Context(
            self.api_key,
            self.api_root,
            self.host,
            auth_key
        )

    @property
    def ctx(self):
        # if the session expired, then reauth
        # and create a new context
        if self._ctx.session_expired:
            self._ctx = self._init_ctx()

        return self._ctx

    def get_root_siren(self):
        return self.ctx.memoized(self.ctx.get_root_siren)


def _get_auth_key(api_key, api_root, host=None):
    key = api_key
    auth_header = "Bearer {}".format(key)
    start_session_url = "{}/start-session".format(api_root)  # TODO: Siren

    session = requests.Session()
    headers = {"Authorization": auth_header}

    # if a host has been provided, then we need to set it on the header
    # and activate the certificate check against the header itself
    # rather than the ip address
    if host:
        session.mount('https://', host_header_ssl.HostHeaderSSLAdapter())
        headers["Host"] = host

    r = session.post(start_session_url, headers=headers)
    if r.status_code == 403:
        api_key_url = api_root.replace('__api', 'dashboard/my-api-keys')
        error_message = "%s Catalog API Key URL: %s" % (AuthenticationFailure.GENERIC_ERROR_MESSAGE, api_key_url)
        raise AuthenticationFailure(message=error_message)
    elif r.status_code != 200:
        raise AuthenticationFailure(response=r)
    
    return r.text



from backports import tempfile
from tests.common import SdkIntegrationTestCase


class SDKSamples(SdkIntegrationTestCase):
    """
    This spec is just meant to catch issues in terms of signature changes
    all the methods are included in one way or another in the SDK documentation
    as samples.

    Avoid including assertions or similar as these tests are not replacements
    for unit or integration tests on these functions.
    """

    def test_get_datasets_in_package(self):
        client = self.client
        package_id = self.create_package("test_get_datasets_in_package")
        builder = self.dataset_builder(package_id, "dataset_for_test_get_dataset_details").with_external_storage(location="jdbc://connectionstring:1232/my-db")
        dataset = client.register_dataset(builder)

        # doc-start
        # returns list of datasets ordered by ascending name
        datasets = client.get_package_datasets(package_id)

        print("Retrieved {} datasets.".format(len(datasets)))
        for ds in datasets:
            print(ds.dataset_id, ds.name)

    def test_get_dataset_details(self):
        client = self.client
        package_id = self.create_package("package_for_test_get_dataset_details")
        builder = self.dataset_builder(package_id, "dataset_for_test_get_dataset_details").with_external_storage(location="jdbc://connectionstring:1232/my-db")
        dataset = client.register_dataset(builder)
        dataset_id = dataset.dataset_id

        # doc-start
        # given a `dataset_id`, dataset metadata can be retrieved as 
        dataset = client.get_dataset(dataset_id)

    def test_get_datafiles_in_dataset(self):
        client = self.client
        package_id = self.create_package("package_for_test_get_datafiles_in_dataset")
        builder = self.dataset_builder(package_id, "dataset_for_test_get_datafiles_in_dataset").with_external_storage(location="jdbc://connectionstring:1232/my-db")
        dataset = client.register_dataset(builder)
        dataset_id = dataset.dataset_id

        files = [{'path': "/path/to/file/A", 'size': 99999}, {'path': "/path/to/file/B", 'size': 88888}]
        datafile = client.register_datafile_metadata(
            dataset_id,
            "datafile_for_test_get_datafiles_in_dataset",
            files,
            '2000-01-01'
        )

        # doc-start
        # returns list of datafiles ordered by descending creation date
        datafiles = client.get_datafiles(dataset_id)

        print("Retrieved {} datafiles.".format(len(datafiles)))
        for df in datafiles:
            print((df.datafile_id, df.files))

    def test_download_datafile(self):
        client = self.client
        package_id = self.create_package("package_for_test_download_datafile")
        dataset = self.register_s3_dataset(package_id, "dataset_for_test_download_datafile", "external-bucket", prefix="samples")
        dataset_id = dataset.dataset_id

        datafile = client.register_s3_datafile(
            dataset_id,
            "datafile_for_test_download_datafile",
            [
                '../test_sandbox/samples/data/AAPL.csv',
                '../test_sandbox/samples/data/MSFT.csv'
            ],
            "samples/",
            '2000-01-01'
        )
        datafile_id = datafile.datafile_id

        with tempfile.TemporaryDirectory() as destination:
            # doc-start
            client.download_datafile(datafile_id, destination)

            import os
            os.listdir(destination)  # shows: ['AAPL.csv', 'MSFT.csv']

    def test_create_datafile_in_dl_s3_storage(self):
        client = self.client
        
        package_id = self.create_package("package_for_test_create_datafile_in_dl_s3_storage")
        dataset = self.register_s3_dataset(package_id, "dataset_for_test_create_datafile_in_dl_s3_storage", "external-bucket")
        dataset_id = dataset.dataset_id

        # doc-start
        # given a dataset with data lake managed storage
        # we can register and upload files into S3
        # which then can be used by your consumers
        datafile = client.register_s3_datafile(
            dataset_id,
            "My Datafile",
            files=[
                '../test_sandbox/samples/data/AAPL.csv',
                '../test_sandbox/samples/data/MSFT.csv'
            ],
            s3_prefix="prefix/",
            data_as_of='2000-01-01'
        )

    def test_create_datafile_in_external_s3_storage(self):
        client = self.client
        package_id = self.create_package("package_for_test_create_datafile_in_external_s3_storage")
        dataset = self.register_s3_dataset(package_id, "dataset_for_test_create_datafile_in_external_s3_storage", "external-bucket", prefix="path")
        dataset_id = dataset.dataset_id

        # doc-start
        # given a dataset with external storage, we want to register a new datafile under it
        # in this case the bucket `external-bucket` is not managed by the data lake, so the upload
        # needs to be done manually
        datafile = client.register_datafile_metadata(
            dataset_id,
            name="test",
            files=[
                {"path": "s3://external-bucket/path/to/file/A", "size": 10000},
                {"path": "s3://external-bucket/path/to/file/B", "size": 15222}
            ],
            data_as_of='2000-01-01'
        )

    def test_update_and_delete_datafile(self):
        client = self.client
        package_id = self.create_package("package_for_test_update_and_delete_datafile")
        builder = self.dataset_builder(package_id, "dataset_for_test_update_and_delete_datafile").with_external_storage(location="jdbc://connectionstring:1232/my-db")
        dataset = client.register_dataset(builder)
        dataset_id = dataset.dataset_id

        files = [{'path': "/path/to/file/A", 'size': 99999}, {'path': "/path/to/file/B", 'size': 88888}]
        datafile = client.register_datafile_metadata(
            dataset_id,
            "datafile_for_test_update_and_delete_datafile",
            files,
            '2000-01-01'
        )

        # doc-start
        # now that we have created a datafile, and have an id (under `datafile.datafile_id`)
        # we can make changes to the metadata
        # this function for example, would change the name
        # while leaving all other attributes as they are.
        updated = client.edit_datafile_metadata(
            datafile.datafile_id,
            name="Updating Datafile."
        )

        # we can also mark the datafile as deleted
        # if we don't want it to be available for consumption anymore.
        # this call only deletes the metadata and leaves
        # the actual data intact.
        client.delete_datafile(updated.datafile_id)



import re
import six
from collections import namedtuple


# regex to convert from camelCase to snake_case
_reg = re.compile(r'(?!^)(?<!_)([A-Z])')


def siren_to_dict(o):
    return {
        attr: getattr(o, attr)
        for attr in dir(o)
        if attr[:1] != '_' and not callable(getattr(o, attr))
    }


def siren_to_entity(o):
    """
    Helper method that converts a siren entity into a namedtuple
    """
    def value_to_entity(v):
        return siren_to_entity(v) if isinstance(v, dict) else v

    # pypermedia does not do recursive translation
    # so we might get a mix of objects or dicts in here
    attrs = siren_to_dict(o) if not isinstance(o, dict) else o
    attrs = {
        to_snake_case(key): value_to_entity(value)
        for key, value in six.iteritems(attrs)
    }
    return namedtuple(o.__class__.__name__, sorted(attrs))(**attrs)


def to_snake_case(s):
    # need to sanitise the string as in some cases the key might look like
    # 'Data Access' or 'Data Sensitivity'
    return _reg.sub(r'_\1', s.replace(' ', '_')).lower()


import logging

from tests.common import SdkIntegrationTestCase, eventually


logger = logging.getLogger(__name__)


class SearchFunctionsTestCase(SdkIntegrationTestCase):

    def entities_are_returned_in_search(self, search_func, expected_min_count):
        def search():
            result = search_func()
            self.assertTrue(len(result) >= expected_min_count)
            return result
        
        return eventually(search)

    def test_search_all_raises_error_for_invalid_count(self):
        self.assert_page_count_is_valid_for_paginated_resource_actions(lambda c: self.client.search(term='searchable', count=c))

    def test_search_all(self):
        self.assertEqual(len(self.client.search(term="")), 0)

        self.create_package("searchable package")
        self.create_collection("searchable collection")

        search_term = "searchable"
        results = self.entities_are_returned_in_search(lambda: self.client.search(search_term), 2)
        entity_types = [e.type for e in results]
        self.assertEqual(any([t == 'package' for t in entity_types]), True)
        self.assertEqual(any([t == 'collection' for t in entity_types]), True)
        
    def test_search_packages_raises_error_for_invalid_count(self):
        self.assert_page_count_is_valid_for_paginated_resource_actions(lambda c: self.client.search_packages(term='searchable', count=c))

    def test_search_packages(self):
        self.assertEqual(len(self.client.search_packages(term="")), 0)

        self.create_package("searchable package")
        search_term = "searchable"

        results = self.entities_are_returned_in_search(lambda: self.client.search_packages(search_term), 1)
        self.assertTrue(all([hasattr(package, "package_id") and search_term in package.name for package in results]))

    def test_search_collections_raises_error_for_invalid_count(self):
        self.assert_page_count_is_valid_for_paginated_resource_actions(lambda c: self.client.search_collections(term='searchable', count=c))
        
    def test_search_collections(self):
        self.assertEqual(len(self.client.search_collections(term="")), 0)

        self.create_collection("searchable collection")
        search_term = "searchable"

        results = self.entities_are_returned_in_search(lambda: self.client.search_collections(search_term), 1)
        self.assertTrue(all([hasattr(collection, "collection_id") and search_term in collection.name for collection in results]))

    def test_search_datasets_raises_error_for_invalid_count(self):
        self.assert_page_count_is_valid_for_paginated_resource_actions(lambda c: self.client.search_datasets(term='searchable', count=c))
        
    def test_search_datasets(self):
        self.assertEqual(len(self.client.search_datasets(term="")), 0)

        package_id = self.create_package("package")
        builder = self.dataset_builder(package_id, "searchable dataset")
        builder = builder.with_external_storage(location="jdbc://somewhere")
        self.client.register_dataset(builder)

        search_term = "searchable"
        results = self.entities_are_returned_in_search(lambda: self.client.search_datasets(search_term), 1)
        for result in results:
            self.assertEqual(result.type, "dataset")
            self.assertIn(search_term, result.name)



from dli.siren import siren_to_entity
from dli.client.utils import ensure_count_is_valid

"""Functions related to the current logged in user."""


class MeFunctions:

    @property
    def __root(self):
        return self.ctx.memoized(
            self.get_root_siren().me,
            "me_root"
        )

    def _get_my_entities(self, entity, my_entities_func, count):
        ensure_count_is_valid(count)

        result = my_entities_func(page_size=count)
        return result.get_entities(rel=entity)

    def get_my_packages(self, count=100):
        """
        Returns a list of packages where session user account is: a Manager, Tech Data Ops or Access Manager.

        :param int count: The number of items to retrieve, defaults to 100.

        :returns: List of my packages.
        :rtype: list[collections.namedtuple]

        - **Sample**

        .. code-block:: python

                my_packages = client.get_my_packages()

        """
        packages = self._get_my_entities('package', self.__root.my_packages, count)
        return [siren_to_entity(p) for p in packages]

    def get_my_consumed_packages(self, count=100):
        """
        Returns a list of all the packages user session account has access to.

        :param int count: The number of items to retrieve, defaults to 100.

        :returns: List of my consumed packages.
        :rtype: list[collections.namedtuple]

        - **Sample**

        .. code-block:: python

                my_consumed_packages = client.get_my_consumed_packages()

        """
        access_requests = self._get_my_entities('access_request', self.__root.list_consumed_packages, count)
        return [siren_to_entity(r.package()) for r in access_requests]

    def get_my_collections(self, count=100):
        """
        Returns a list of collections where session user account is the manager.

        :param int count: The number of items to retrieve, defaults to 100.

        :returns: List of my collections.
        :rtype: list[collections.namedtuple]

        - **Sample**

        .. code-block:: python

                my_collections = client.get_my_collections()

        """
        collections = self._get_my_entities('collection', self.__root.my_collections, count)
        return [siren_to_entity(c) for c in collections]

    def get_my_accounts(self):
        """
        Returns a list of all the accounts associated with user session.

        :returns: list of all associated accounts.
        :rtype: list[collections.namedtuple]

        - **Sample**

        .. code-block:: python

                my_accounts = client.get_my_accounts()

        """
        result = self.get_root_siren().list_my_accounts()
        accounts = result.get_entities(rel="")
        return [siren_to_entity(a) for a in accounts]



import datetime
import json
import logging
import os
import socket
import time
import unittest
import uuid

import requests
import s3fs
from mock import patch
from six.moves.urllib import parse

from dli.client import builders, session
from tests import localstack_helper

logger = logging.getLogger(__name__)
# this token allows us to login when running datacat in dev mode
token = "eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiIsImlhdCI6MTUxMzMzMTUwOSwiZXhwIjo5NTEzMzM1MTA5fQ.eyJhdWQiOiJkYXRhbGFrZS1hY2NvdW50cyIsImF1dGhfdGltZSI6MTUxMzMzMTUwOSwiZGF0YWxha2UiOnsiYWNjb3VudHMiOnsiaWJveHgiOiJydyIsIm1yZCI6InIiLCJkYXRhbGFrZS1tZ210IjoicncifX0sImVtYWlsIjoiamltQGV4YW1wbGUuY29tIiwiZW1haWxfdmVyaWZpZWQiOnRydWUsImV4cCI6OTUxMzMzNTEwOSwiaWF0IjoxNTEzMzMxNTA5LCJpc3MiOiJodHRwczovL2h5ZHJhLWRldi51ZHBtYXJraXQubmV0Iiwibm9uY2UiOiI1YTgyMzI2NC1iN2QzLTQ4NWUtYjc2MS1lYzE1MTI5NDQ4MTQiLCJzdWIiOiJ1c2VyOjEyMzQ1OmppbSJ9.mB9-W8KCyARq5zkLsjSdmE4KKQ7Q7MBvmX5J6rYMMtM"


def build_fake_s3fs(key, secret, token):
    # Nasty: Resolve the hostname so that the we connect
    # using the ip address instead of the hostname
    # this is due a limitation in fake-s3, read more here:
    # https://github.com/jubos/fake-s3/issues/17
    s3_hostname = socket.gethostbyname(
        os.environ.get("FAKE_S3_HOST", "localhost"))
    s3_port = os.environ.get("FAKE_S3_PORT", 4569)

    return s3fs.S3FileSystem(
        key=key,
        secret=secret,
        token=token,
        client_kwargs={
            "endpoint_url": "http://%s:%s" % (s3_hostname, s3_port)
        }
    )


class SdkIntegrationTestCase(unittest.TestCase):
    """
    Helper TestCase to test against a local docker container running the datacat api.
    To run these locally, run `docker-compose up` on the root directory
    """

    aws_account_id = "12345"

    # FIXME
    # -----
    # By now we patch the response to return a token we know will authenticate
    # instead of generating an api key, this isn't ideal as the behaviour is
    # not exactly the same, but it is good enough by now.
    @patch("dli.client.dli_client._get_auth_key", lambda k, r, h: token)
    def setUp(self):
        self.headers = {
            'Content-type': 'application/json',
            'Authorization': 'Bearer {}'.format(token),
            'Cookie': 'oidc_id_token=' + token
        }
        self.root_url = os.environ.get(
            "DATA_LAKE_INTERFACE_URL", "http://localhost:9000/__api/")
        self.api_key = self.get_api_key()
        self.client = self.create_session()
        self.s3 = []

        # create a fake s3 repo
        self.set_s3_client_mock()
        self.register_aws_account(self.aws_account_id)
        self.s3_client = localstack_helper.get_s3_client()

    def set_s3_client_mock(self):
        def _upload(_, files, location, tr=None, r=None):
            result = []
            for f in files:
                path = location + os.path.basename(f)
                self.s3.append(path)
                result.append({"path": "s3://" + path})
            return result

        s3_upload = patch('dli.client.s3.Client.upload_files_to_s3', _upload)
        s3_upload.start()
        s3_download = patch('dli.client.s3.Client.download_files_from_s3_path')
        self.s3_download_mock = s3_download.start()

        # cleanup
        self.addCleanup(s3_upload.stop)
        self.addCleanup(s3_download.stop)

    def get_api_key(self):
        # url = "%s/generate-my-key" % self.root_url
        # payload = {
        #     "account": "datalake-mgmt",
        #     "rights": "rw",
        #     "expiration": "2030-01-01"
        # }

        # response = requests.post(url, data=json.dumps(payload), headers=self.headers)
        # return response.text
        return "key"

    def create_package(self, name, access="Restricted", **kwargs):
        payload = {
            "name": name + "_" + str(uuid.uuid4()),
            "description":  "asd",
            "publisher":  "datalake-mgmt",
            "manager_id":  "datalake-mgmt",
            "access_manager_id":  "datalake-mgmt",
            "tech_data_ops_id":  "datalake-mgmt",
            "topic":  "Academic/Education",
            "access":  access,
            "internal_data":  "No",
            "data_sensitivity":  "Public",
            "terms_and_conditions":  "Terms and Conditions"
        }
        payload.update(**kwargs)
        return self.client.register_package(**payload).package_id

    def create_collection(self, name):
        return self.client.create_collection(
            name=name + "_" + str(uuid.uuid4()),
            description="test",
            manager_id="datalake-mgmt",
        ).collection_id

    def dataset_builder(self, package_id, dataset_name, **kwargs):
        return builders.DatasetBuilder(
            package_id=package_id,
            name=dataset_name,
            description="a testing dataset",
            content_type="Pricing",
            data_format="CSV",
            publishing_frequency="Daily",
            taxonomy=[],
            **kwargs
        )

    def register_s3_dataset(self, package_id, dataset_name, bucket_name, prefix="prefix"):
        self._setup_bucket_and_prefix(bucket_name, prefix)
        return self.client.register_dataset(
            self.dataset_builder(package_id, dataset_name).with_external_s3_storage(
                bucket_name,
                self.aws_account_id,
                prefix
            )
        )

    def _setup_bucket_and_prefix(self, bucket_name, prefix):
        # API checks whether DataLake can access S3 bucket at prefix. If dataset is registered with S3 location
        # we must create a fake one in localstack.
        s3_client = self.s3_client
        s3_client.create_bucket(Bucket=bucket_name)
        s3_client.put_object(Bucket=bucket_name, Key=prefix, Body="")

        def cleanup_bucket():
            for object_to_delete in s3_client.list_objects(Bucket=bucket_name)['Contents']:
                s3_client.delete_objects(Bucket=bucket_name, Delete={"Objects": [{"Key": object_to_delete['Key']}]})
            s3_client.delete_bucket(Bucket=bucket_name)

        self.addCleanup(cleanup_bucket)


    def create_session(self):
        return session.start_session(self.api_key, self.root_url)

    def register_aws_account(self, aws_account_id):
        response = requests.post(
            parse.urljoin(self.root_url, "me/aws-accounts"),
            headers=self.headers,
            data=json.dumps({
                "awsAccountId": str(aws_account_id),
                "awsAccountName": str(aws_account_id),
                "awsAccessKeyId": "accessKey",
                "awsSecretAccessKey": "secretAccessKey",
                "awsRegion": "eu-west-1",
                "accountIds": ["datalake-mgmt"]
            })
        )

        self.assertEqual(response.status_code, 200)

    def assert_page_count_is_valid_for_paginated_resource_actions(self, func):
        with self.assertRaises(ValueError):
            func(-1)
        with self.assertRaises(ValueError):
            func(0)
        with self.assertRaises(ValueError):
            func("test")


def eventually(assertion, delay=1, retries=5):
    try:
        return assertion()
    except Exception as e:
        if retries <= 1:
            raise e
        time.sleep(delay)
        return eventually(assertion, delay=delay, retries=retries-1)



import boto3
import json
import os


LOCALSTACK_DUMMY_AWS_ACCOUNT_ID = "123456789012"
LOCALSTACK_AWS_REGION = 'us-east-1'


def _fake_aws_credentials():
    return {
        'region_name': LOCALSTACK_AWS_REGION,
        'aws_access_key_id': 'dummy-access-key',
        'aws_secret_access_key': 'dummy-secret',
        'aws_session_token': 'dummy-token'
    }


def get_s3_client():
    s3_localstack_url = os.environ.get("S3_LOCALSTACK_URL", 'http://localstack:4572')
    return boto3.client(
        's3',
        endpoint_url=s3_localstack_url,
        **_fake_aws_credentials()
    )


def get_sns_client():
    sns_localstack_url = os.environ.get("SNS_LOCALSTACK_URL", 'http://localstack:4575')
    return boto3.client(
        'sns',
        endpoint_url=sns_localstack_url,
        **_fake_aws_credentials()
    )


def get_sqs_client():
    sqs_localstack_url = os.environ.get("SQS_LOCALSTACK_URL", 'http://localstack:4576')
    return boto3.client(
        'sqs',
        endpoint_url=sqs_localstack_url,
        **_fake_aws_credentials()
    )


def get_topic_arn(topic_name):
    return "arn:aws:sns:%s:%s:%s" \
            % (LOCALSTACK_AWS_REGION, LOCALSTACK_DUMMY_AWS_ACCOUNT_ID, topic_name)


import requests

from urllib.parse import urlparse, urljoin
from dli import __version__
import json as json_lib
import time

from dli.client.exceptions import (
    InvalidPayloadException,
    UnAuthorisedAccessException,
    InsufficientPrivilegesException,
    DatalakeException
)

# Python 2.7 / 3.x differences
try:
    from json.decoder import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError


def make_hook(root, get_header):
    def _make_empty_response(r):
        import copy
        response = copy.deepcopy(r)
        response._content = b'{"class": ["none"]}'
        return response

    def _extract_error_response_message(r):
        try:
            return r.json()['errorText']
        except (JSONDecodeError, KeyError):
            return r.text

    def _response_hook(r, *args, **kwargs):

        if r.status_code in [201, 202, 204, 404]:  # for now. or (300 <= r.status_code <= 399):
            return _make_empty_response(r)
        elif r.status_code in [400, 422]:
            raise InvalidPayloadException(_extract_error_response_message(r))
        elif r.status_code == 401:
            raise UnAuthorisedAccessException(_extract_error_response_message(r))
        elif r.status_code == 403:
            raise InsufficientPrivilegesException(_extract_error_response_message(r))
        elif r.status_code > 400:
            raise DatalakeException(_extract_error_response_message(r))
    return _response_hook


class DliRequestFactoryFactory(object):
    def __init__(self, root, host=None, auth_header=lambda: {}):
        self.root = root
        self.host = host
        self.auth_header = auth_header

    def request_factory(self, method=None, url=None, headers=None, files=None, data=None,
                        params=None, auth=None, cookies=None, hooks=None, json=None):

        # relative uri? make it absolute.
        if not urlparse(url).netloc:
            url = urljoin(str(self.root), str(url))     # python 2/3 nonsense

        # uri template substitution.
        pars = params if method == "GET" else data

        if pars:
            if '__json' in pars:
                js = json_lib.loads(pars['__json'])
                pars.update(js)
                del pars['__json']

        if pars and method == "GET":
            params = pars
        else:
            json = pars
            data = None

        # populate headers
        if not headers:
            headers = {}
        headers['Content-Type'] = "application/vnd.siren+json"
        headers["X-Data-Lake-SDK-Version"] = str(__version__)
        headers.update(self.auth_header())

        # if a host has been provided, then we need to set it on the header
        if self.host:
            headers['Host'] = self.host

        if not hooks:
            hooks = {}
        if 'response' not in hooks:
            hooks['response'] = []
        hooks['response'] = make_hook(self.root, self.auth_header)

        return requests.Request(
            method=method,
            url=url,
            headers=headers,
            files=files,
            data=data,
            params=params,
            auth=auth,
            cookies=cookies,
            hooks=hooks,
            json=json
        )



import unittest
from dli.client.context import Context


class ContextTestCase(unittest.TestCase):

    valid_token = "eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiIsImlhdCI6MTUxMzMzMTUwOSwiZXhwIjo5NTEzMzM1MTA5fQ.e30.fjav6VOdXAnE6WgUwSxEbrNuiqQBSEalOd177F9UrCM"
    expired_token = "eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiIsImlhdCI6MTUxMzMzMTUwOSwiZXhwIjoxNTEzMzM1MTA5fQ.e30.-vTTDvptpjPrG2PO1GBo1OOlBqvx6D_gPGnU7Tw-YII"

    def test_can_decode_valid_jwt_token(self):
        ctx = Context(
            "key",
            "root",
            None,
            self.valid_token
        )
        self.assertFalse(ctx.session_expired)

    def test_can_detect_token_is_expired(self):
        ctx = Context(
            "key",
            "root",
            None,
            self.expired_token
        )
        self.assertTrue(ctx.session_expired)

    def test_when_token_cant_be_decoded_then_we_assume_no_session_expiration(self):
        ctx = Context(
            "key",
            "root",
            None,
            "invalid.token"
        )
        self.assertFalse(ctx.session_expired)


import boto3
import logging
import six

from .common import SdkIntegrationTestCase
from dli.client import builders
from dli.client.exceptions import (
    CatalogueEntityNotFoundException,
    MissingMandatoryArgumentException,
    InvalidPayloadException,
)

from . import localstack_helper


class AutoRegMetadataFunctionsTestCase(SdkIntegrationTestCase):

    test_topic = 's3-notifications'
    test_bucket = 'test-bucket'

    s3 = localstack_helper.get_s3_client()
    sns = localstack_helper.get_sns_client()

    @classmethod
    def setUpClass(cls):
        cls.sns.create_topic(Name=cls.test_topic)

    @classmethod
    def tearDownClass(cls):
        cls.sns.delete_topic(TopicArn=localstack_helper.get_topic_arn(cls.test_topic))

    def setUp(self):
        super(AutoRegMetadataFunctionsTestCase, self).setUp()

        self.package_id = self.create_package("test_auto_reg_metadata_functions")
        self.dataset = self.register_s3_dataset(
                self.package_id,
                "test_auto_reg_metadata_functions",
                self.test_bucket,
                "test/auto")

    def create_test_auto_reg_metadata(self, **kwargs):
        params = {
            'dataset_id': self.dataset.dataset_id,
            'path_template': 'path/to/as_of_date={{ year }}-{{ month }}-{{ day }}/type={{ type }}',
            'name_template': "Datafile_{{ type }}_{{ year }}-{{ month }}-{{ day }}",
            'as_of_date_template': "{{ year }}-{{ month }}-{{ day }}",
            'active': True,
            'sns_topic_for_s3_events': self.test_topic,
            'handle_files': False
        }

        if kwargs:
            params.update(**kwargs)

        return self.client.set_auto_registration_metadata(
            dataset_id=params['dataset_id'],
            path_template=params['path_template'],
            name_template=params['name_template'],
            as_of_date_template=params['as_of_date_template'],
            active=params['active'],
            sns_topic_for_s3_events=params['sns_topic_for_s3_events'],
            handle_files=params['handle_files']
        )

    def test_cannot_register_auto_reg_metadata_if_missing_dataset(self):
        with self.assertRaises(MissingMandatoryArgumentException):
            self.create_test_auto_reg_metadata(dataset_id=None)
        with self.assertRaises(MissingMandatoryArgumentException):
            self.create_test_auto_reg_metadata(dataset_id='')

    def test_cannot_register_auto_reg_metadata_for_unknown_dataset(self):
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.create_test_auto_reg_metadata(dataset_id='unknown')

    def assert_cannot_register_auto_reg_metadata_if_missing_mandatory_parameter(self, param_name):
        with self.assertRaises(InvalidPayloadException):
            self.create_test_auto_reg_metadata(**{param_name: None})
        with self.assertRaises(InvalidPayloadException):
            self.create_test_auto_reg_metadata(**{param_name: ''})

    def test_cannot_register_auto_reg_metadata_if_missing_path_template(self):
        self.assert_cannot_register_auto_reg_metadata_if_missing_mandatory_parameter('path_template')

    def test_cannot_register_auto_reg_metadata_if_missing_name_template(self):
        self.assert_cannot_register_auto_reg_metadata_if_missing_mandatory_parameter('name_template')

    def test_cannot_register_auto_reg_metadata_if_missing_as_of_date_template(self):
        self.assert_cannot_register_auto_reg_metadata_if_missing_mandatory_parameter('as_of_date_template')

    def test_cannot_register_auto_reg_metadata_if_missing_active_status_flag(self):
        self.assert_cannot_register_auto_reg_metadata_if_missing_mandatory_parameter('active')

    def test_cannot_register_auto_reg_metadata_if_undefined_variables_in_name_template(self):
        path_template = 'path/to/as_of_date={{ year }}-{{ month }}-{{ day }}/type={{ type }}'
        name_template = "Datafile_{{ some_undefined }}_{{ year }}-{{ month }}-{{ day }}"
        with self.assertRaises(InvalidPayloadException):
            self.create_test_auto_reg_metadata(path_template=path_template, name_template=name_template)

    def test_cannot_register_auto_reg_metadata_if_undefined_variables_in_as_of_date_template(self):
        path_template = 'path/to/as_of_date={{ year }}-{{ month }}-{{ day }}/type={{ type }}'
        as_of_date_template = "{{ some_undefined }}_{{ year }}-{{ month }}-{{ day }}"
        with self.assertRaises(InvalidPayloadException):
            self.create_test_auto_reg_metadata(path_template=path_template, as_of_date_template=as_of_date_template)

    def test_cannot_register_auto_reg_metadata_if_unused_variables_in_path_template(self):
        path_template = 'path/to/{{ unused }}/as_of_date={{ year }}-{{ month }}-{{ day }}/type={{ type }}'
        name_template = "Datafile_{{ year }}-{{ month }}-{{ day }}"
        as_of_date_template = "{{ year }}-{{ month }}-{{ day }}"
        with self.assertRaises(InvalidPayloadException):
            self.create_test_auto_reg_metadata(path_template=path_template, name_template=name_template, as_of_date_template=as_of_date_template)

    def test_cannot_register_auto_reg_metadata_for_non_s3_dataset(self):
        dataset = self.client.register_dataset(
            self.dataset_builder(
                self.package_id,
                "test_cannot_register_auto_reg_metadata_for_non_s3_dataset"
            ).with_external_storage(location="jdbc://connectionstring:1232/my-db")
        )
        with self.assertRaises(InvalidPayloadException):
            self.create_test_auto_reg_metadata(dataset_id=dataset.dataset_id)

    def test_can_register_auto_reg_metadata_for_dataset(self):
        auto_reg_metadata = self.create_test_auto_reg_metadata()
        self.assertIsNotNone(auto_reg_metadata)
        self.assertEqual(auto_reg_metadata.dataset_id, self.dataset.dataset_id)

    def test_can_edit_has_auto_reg_failure_flag(self):
        auto_reg_metadata = self.create_test_auto_reg_metadata(has_auto_registry_failure_notifications=True)
        edited_auto_reg_metadata = self.client.edit_auto_registration_metadata(
            auto_reg_metadata.auto_reg_metadata_id, has_auto_registry_failure_notifications=False
        )
        
        assert auto_reg_metadata.has_auto_registry_failure_notifications
        assert not edited_auto_reg_metadata.has_auto_registry_failure_notifications

    def test_has_auto_reg_failure_flag_should_not_change_when_set_to_none(self):
        auto_reg_metadata = self.create_test_auto_reg_metadata(has_auto_registry_failure_notifications=True)
        auto_reg_metadata = self.client.edit_auto_registration_metadata(
            auto_reg_metadata.auto_reg_metadata_id,
            has_auto_registry_failure_notifications=None
        )
        assert auto_reg_metadata.has_auto_registry_failure_notifications

    def test_cannot_get_auto_reg_metadata_if_mising_dataset_id(self):
        with self.assertRaises(ValueError):
            self.client.get_auto_registration_metadata(dataset_id='')
        with self.assertRaises(ValueError):
            self.client.get_auto_registration_metadata(dataset_id=None)

    def test_cannot_get_auto_reg_metadata_if_mising_auto_reg_metadata_id(self):
        with self.assertRaises(ValueError):
            self.client.get_auto_registration_metadata(auto_reg_metadata_id='')
        with self.assertRaises(ValueError):
            self.client.get_auto_registration_metadata(auto_reg_metadata_id=None)

    def test_cannot_get_auto_reg_metadata_if_mising_dataset_id_and_auto_reg_metadata_id(self):
        with self.assertRaises(ValueError):
            self.client.get_auto_registration_metadata(dataset_id='', auto_reg_metadata_id='')

    def test_cannot_get_auto_reg_metadata_for_unknown_dataset_id(self):
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.get_auto_registration_metadata(dataset_id='unknown')

    def test_cannot_get_auto_reg_metadata_for_unknown_auto_reg_metadata_id(self):
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.get_auto_registration_metadata(auto_reg_metadata_id='unknown')

    def test_cannot_get_auto_reg_metadata_for_unknown_dataset_id_and_auto_reg_metadata_id(self):
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.get_auto_registration_metadata(dataset_id='unknown', auto_reg_metadata_id='unknown')

    def test_can_get_auto_reg_metadata_by_dataset_id(self):
        auto_reg_metadata = self.create_test_auto_reg_metadata()
        result = self.client.get_auto_registration_metadata(dataset_id=auto_reg_metadata.dataset_id)
        self.assertIsNotNone(result)
        self.assertEqual(result.auto_reg_metadata_id, auto_reg_metadata.auto_reg_metadata_id)

    def test_can_get_auto_reg_metadata_by_auto_reg_metadata_id(self):
        auto_reg_metadata = self.create_test_auto_reg_metadata()
        result = self.client.get_auto_registration_metadata(auto_reg_metadata_id=auto_reg_metadata.auto_reg_metadata_id)
        self.assertIsNotNone(result)
        self.assertEqual(result.auto_reg_metadata_id, auto_reg_metadata.auto_reg_metadata_id)

    def test_can_get_auto_reg_metadata_by_dataset_id_and_auto_reg_metadata_id(self):
        auto_reg_metadata = self.create_test_auto_reg_metadata()
        result = self.client.get_auto_registration_metadata(
                    dataset_id=auto_reg_metadata.dataset_id,
                    auto_reg_metadata_id=auto_reg_metadata.auto_reg_metadata_id
                )
        self.assertIsNotNone(result)
        self.assertEqual(result.auto_reg_metadata_id, auto_reg_metadata.auto_reg_metadata_id)

    def test_cannot_delete_auto_reg_metadata_if_missing_auto_reg_id(self):
        with self.assertRaises(MissingMandatoryArgumentException):
            self.client.delete_auto_registration_metadata(None)
        with self.assertRaises(MissingMandatoryArgumentException):
            self.client.delete_auto_registration_metadata('')

    def test_cannot_delete_unknown_auto_reg_metadata(self):
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.delete_auto_registration_metadata("unknown")

    def test_can_delete_auto_reg_metadata(self):
        auto_reg_metadata = self.create_test_auto_reg_metadata()
        # Delete 
        self.client.delete_auto_registration_metadata(auto_reg_metadata.auto_reg_metadata_id)
        # Try fetching the same will throw error now
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.get_auto_registration_metadata(auto_reg_metadata_id=auto_reg_metadata.auto_reg_metadata_id)

    def test_cannot_edit_auto_reg_metadata_if_missing_auto_reg_id(self):
        with self.assertRaises(MissingMandatoryArgumentException):
            self.client.edit_auto_registration_metadata(None)
        with self.assertRaises(MissingMandatoryArgumentException):
            self.client.edit_auto_registration_metadata('')

    def test_cannot_edit_unknown_auto_reg_metadata(self):
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.edit_auto_registration_metadata("unknown")

    def test_can_edit_auto_reg_metadata(self):
        auto_reg_metadata = self.create_test_auto_reg_metadata()
        self.assertEqual(auto_reg_metadata.active, True)

        # Edit the metadata. Make it inactive for example
        edited_metadata = self.client.edit_auto_registration_metadata(
            auto_reg_metadata.auto_reg_metadata_id,
            active=False            
        )
        self.assertEqual(edited_metadata.active, False)


import logging
import six

from dli.client.exceptions import (
    CatalogueEntityNotFoundException,
    MissingMandatoryArgumentException,
)
from dli.siren import siren_to_entity, siren_to_dict
from dli.client.utils import ensure_count_is_valid, filter_out_unknown_keys, to_camel_cased_dict

from dli.client.builders import DatasetLocationBuilder

logger = logging.getLogger(__name__)


class DatasetFunctions:
    _KNOWN_FIELDS = set([
        "location",
        "packageId",
        "name",
        "description",
        "keywords",
        "contentType",
        "location",
        "dataFormat",
        "publishingFrequency",
        "namingConvention",
        "documentation",
        "taxonomy",
        "loadType"
    ])
    @property
    def __root(self):
        return self.ctx.memoized(
            self.get_root_siren().datasets_root,
            "datasets_root"
        )

    def get_s3_access_keys_for_dataset(self, *dataset_ids):
        """
        Retrieve S3 access keys for the specified account to access the
        specified dataset(s).

        :param list dataset_ids: One ore more dataset ids to get access to.
        :returns: A namedtuple containing the AWS keys and session token.
        :rtype: collections.namedtuple
        
        - **Sample**

        .. code-block:: python

                s3_access_keys = client.get_s3_access_keys_for_dataset(dataset_id1, dataset_id2)
                # print(s3_access_keys)
                # access_key(access_key_id='39D19A440AFE452B9', secret_access_key='F426A93CDECE45C9BFF8F4F19DA5CB81', session_token='C0CC405803F244CA99999')

        """
        if not dataset_ids:
            raise MissingMandatoryArgumentException('dataset_ids')

        # validate that all datasets exist
        for dataset_id in dataset_ids:
            self._get_dataset(dataset_id=dataset_id)

        payload = {"datasetIds": list(dataset_ids)}
        response = self.get_root_siren().request_access_keys(__json=payload)
        keys = siren_to_entity(response)
        return keys

    def get_dataset(self, id=None, name=None, package_id=None, package_name=None):
        """
        Retrieves a dataset.

        :param str id: The id of the dataset.
        :param str name: The name of the dataset.
        :param str package_id: The id of the package to which this dataset belongs. 
                            Either this or package name is required if dataset is being looked up by name.
        :param str package_name: The name of the package to which this dataset belongs.
                            Either this or package id is required if dataset is being looked up by name.

        :returns: The dataset.
        :rtype: collections.namedtuple

        - **Sample**

        .. code-block:: python

                # Look up by dataset id
                dataset = client.get_dataset('my_dataset_id')
                # or equivalent
                dataset = client.get_dataset(id='my_dataset_id')

                # Look up by dataset name 

                # If package id is known
                dataset = client.get_dataset(name='my_dataset', package_id='my_package_id')

                # if package name is known
                dataset = client.get_dataset(name='my_dataset', package_name='my_package')
        """

        if id is not None:
            return siren_to_entity(self._get_dataset(dataset_id=id))

        if name is not None:
            return siren_to_entity(self._get_dataset(name=name, package_id=package_id, package_name=package_name))

        raise ValueError("Either dataset id or name (along with package id or package name) must be specified to look up dataset")

    def _get_dataset(self, **kwargs):

        dataset = self.__root.get_dataset(**kwargs)

        if not dataset:
            raise CatalogueEntityNotFoundException('Dataset', params=kwargs)

        return dataset

    def register_dataset(self, builder):
        """
        Submit a request to create a new dataset under a specified package in the Data Catalogue.

        :param dli.client.builders.DatasetBuilder builder: An instance of DatasetBuilder. This builder object sets sensible defaults and exposes
                                                           helper methods on how to configure its storage options.

        :returns: A newly created Dataset.
        :rtype: collections.namedtuple

        - **Sample**

        .. code-block:: python

                # Please refer to builder docs for examples on
                # how to create an instance of DatasetBuilder.

                dataset = client.register_dataset(builder)
        """
        payload = builder.build()

        result = self.__root.create_dataset(__json=payload)
        return siren_to_entity(result)

    def edit_dataset(
        self,
        dataset_id,
        location_builder=None,
        **kwargs
    ):
        """
        Updates information on a dataset, returning the updated instance.
        If keyword argument is not specified field keeps its old value.
        Optional enum and text type fields can be unset by passing ``None``.

        :param str dataset_id: Id of the dataset being updated.
        :param dli.client.builders.DatasetLocationBuilder location_builder: Optional. An instance of DatasetLocationBuilder. This builder object exposes
                                                                            helper methods to configure dataset storage options.

        Keyword arguments:

        :keyword str name: Optional. A descriptive name of a dataset. It should be unique within the package.
        :keyword str description: Optional. A short description of a package.
        :keyword str content_type: Optional. A way for the data steward to classify the type of data in the dataset
                                (e.g. pricing).
        :keyword str data_format: Optional. The format of the data: `CSV`, `IMAGE`, `JSON`, `PARQUET`, `TXT`, `XML`, `Other`.
        :keyword str publishing_frequency: Optional. The internal on which data is published. Possible values are: `Hourly`,
                                        `Daily`, `Weekday`, `Weekly`, `Monthly`, `Quarterly`, `Yearly`, `Not Specified`.
        :keyword list[str] keywords: Optional. User-defined list of keywords that can be used to find this dataset through
                                    the search interface.
        :keyword str naming_convention: Optional. Key for how to read the dataset name.
        :keyword str documentation: Optional. Documentation about this dataset in markdown format.
        :keyword list[str] taxonomy: Optional. A list of segments to be used for a taxonomy,
                                    the Data-<< Organization >>-<< topic >> prefix will be applied by the catalogue  For
                                    a taxonomy of Data-IHS Markit-Financial Markets-Credit-CDS, you would provide
                                    `taxonomy=["Credit", "CDS"]`
        :keyword str load_type: Optional. Whether each datafile in this dataset should be considered as a full version
                                of a dataset or a delta or increment. Accepted types are `Full Load`, `Incremental Load`

        :returns: Updated Dataset.
        :rtype: collections.namedtuple

        - **Sample**

        .. code-block:: python

                # e.g. update dataset description
                updated_dataset = client.edit_dataset(
                    "my-dataset-id",
                    description="Updated my dataset description"
                )

                # update dataset location. Please note that this is only allowed if the dataset has no datafiles registered.
                builder = DatasetLocationBuilder().with_external_storage("external-storage-location")
                updated_dataset = client.edit_dataset(
                    "my-dataset-id",
                    location_builder=builder
                )

                # update dataset taxonomy
                updated_dataset = client.edit_dataset(
                    "my-dataset-id",
                    taxonomy=["Credit", "CDS"]
                )

        """
        if not dataset_id:
            raise MissingMandatoryArgumentException('dataset_id')

        dataset = self._get_dataset(dataset_id=dataset_id)
        fields = filter_out_unknown_keys(to_camel_cased_dict(kwargs), DatasetFunctions._KNOWN_FIELDS)
        fields["packageId"] = dataset.packageId

        if isinstance(location_builder, DatasetLocationBuilder):
            fields.update(location_builder.build())
        elif location_builder is None:
            fields["location"] = dataset.location
        else:
            raise TypeError(
                'location_builder must be a subclass of '
                'dli.client.builders.DatasetLocationBuilder '
                'not %s' % type(location_builder)
            )

        # clean up any unknown fields, and update the entity
        dataset_as_dict = siren_to_dict(dataset)
        for key in list(dataset_as_dict.keys()):
            if key not in DatasetFunctions._KNOWN_FIELDS:
                del dataset_as_dict[key]

        dataset_as_dict.update(fields)

        # perform the update and return the resulting entity
        return siren_to_entity(dataset.edit_dataset(__json=dataset_as_dict))

    def delete_dataset(self, dataset_id):
        """
        Marks a particular dataset (and all its datafiles) as deleted.
        This dataset will no longer be accessible by consumers.

        :param str dataset_id: The id of the dataset to be deleted.

        :returns: None

        - **Sample**

        .. code-block:: python

                client.delete_dataset(dataset_id)

        """
        if not dataset_id:
            raise MissingMandatoryArgumentException('dataset_id')

        dataset = self._get_dataset(dataset_id=dataset_id)

        dataset.delete_dataset()

    def get_datafiles(self, dataset_id, name_contains=None, as_of_date_start=None, as_of_date_end=None, count=100):
        """
        Returns a list of all datafiles registered under a dataset.

        :param str dataset_id: The id of the dataset.
        :param str name_contains: Optional. Look up only those datafiles for the dataset where name contains this string.
        :param str as_of_date_start: Optional. Datafiles having data_as_of date greater than or equal to this date. This must be specified in YYYY-MM-DD format.
        :param str as_of_date_end: Optional. Datafiles having data_as_of date less than or equal to this date. This must be specified in YYYY-MM-DD format.
        :param int count: Optional count of datafiles to be returned. Defaults to 100.

        :returns: List of all datafiles registered under the dataset.
        :rtype: list[collections.namedtuple]

        - **Sample**

        .. code-block:: python

                datafiles = client.get_datafiles(
                    dataset_id,
                    name_contains='My Test Data',
                    as_of_date_start='2018-10-11',
                    as_of_date_end='2018-10-15',                    
                    count=10
                )
        """
        if not dataset_id:
            raise MissingMandatoryArgumentException('dataset_id')

        ensure_count_is_valid(count)

        params = {
            'name': name_contains,
            'as_of_date_start': as_of_date_start,
            'as_of_date_end': as_of_date_end,
            'page_size': count
        }

        query_params = {k: v for k, v in six.iteritems(params) if v is not None}

        dataset = self._get_dataset(dataset_id=dataset_id)

        datafiles = dataset.get_datafiles(**query_params).get_entities(rel="datafile")

        return [siren_to_entity(df) for df in datafiles]

    def get_latest_datafile(self, dataset_id):
        """
        Fetches datafile metadata of latest datafile in the dataset.

        :param str dataset_id: The id of the dataset.

        :returns: The datafile.
        :rtype: collections.namedtuple

        - **Sample**

        .. code-block:: python

                datafile = client.get_latest_datafile(dataset_id)
        """
        if not dataset_id:
            raise MissingMandatoryArgumentException('dataset_id')

        dataset = self._get_dataset(dataset_id=dataset_id)
        if not dataset:
            raise CatalogueEntityNotFoundException('Dataset', {'dataset_id': dataset_id})

        datafile = dataset.get_latest_datafile()
        if not datafile:
            raise CatalogueEntityNotFoundException('Datafile', message='Dataset has no Datafiles')

        return siren_to_entity(datafile)

    def _register_schema(self, dataset_id, payload):
        if not dataset_id:
            raise MissingMandatoryArgumentException('dataset_id')

        dataset = self._get_dataset(dataset_id=dataset_id)

        payload = {k: v for k, v in payload.items() if v is not None}
        return siren_to_entity(dataset.create_schema(__json=payload))

    def register_schema(
        self,
        dataset_id,
        version,
        valid_as_of,
        fields,
        type='StructType',
        partitions=None,
        description=None
    ):
        """
        Registers schema metadata for a dataset.

        :param str dataset_id: Id of the dataset for the schema.
        :param str version: A user assigned version name/number. It should be unique within the dataset.
        :param str valid_as_of: The date as of which the schema is active.
                               Expected format is YYYY-MM-DD.
        :param list[dict] fields: Non empty list of `Field` as described below.
        :param str type: Optional. Type of the schema. Defaults to `StructType`.
        :param list[dict] partitions: Optional. Non empty list of `Partition` as described below.
        :param str description: Optional. Description for the schema.

        .. code-block:: python

                # Field: Spark compatible
                {
                    "name": "field_a", 			# name of the column.
                    "type": "String", 			# string representation of the type of the column (might be another StructType if it is a nested type).
                    "nullable": True,  			# defaulted to True - A boolean indicating whether the field is nullable or not.
                    "metadata": None			# optional dictionary with metadata for this column.
                }

                # Partition:
                {
                    "name": "key1",
                    "type": "String"
                }    

        :returns: The registered schema.
        :rtype: collections.namedtuple

        - **Sample**

        .. code-block:: python

                my_schema_fields = [
                            {
                                'name': 'field_1',
                                'type': 'String',
                                'nullable': False
                            },
                            {
                                'name': 'field_2',
                                'type': 'Double',
                                'nullable': False
                            },
                            {
                                'name': 'field_3',
                                'type': 'Int',
                                'nullable': True,
                                'metadata': {
                                    'some_key': 'some_value'
                                }
                            },
                        ]
                my_schema_partitions = [
                    {
                        'name': 'field_1',
                        'type': 'String'
                    }
                ]

                my_schema = client.register_schema(
                    "my-dataset-id",
                    version='1a',
                    valid_as_of='2018-10-31',
                    fields=my_schema_fields,
                    type='my-schema-type',
                    partitions=my_schema_partitions,
                    description="My schema description"
                )
        """

        payload = {
            'version': version,
            'validAsOf': valid_as_of,
            'type': type,
            'fields': fields,
            'partitions': partitions,
            'description': description,
        }

        return self._register_schema(dataset_id, payload)

    def register_spark_schema(
        self,
        dataset_id,
        version,
        valid_as_of,
        spark_schema_json,
        partitions=None,
        description=None
    ):
        """
        Registers dataframe schema for the datalake `dataset` for instances where data ingest pipelines make use of spark.

        :param str dataset_id: Id of the dataset for the schema.
        :param str version: A user assigned version name/number. It should be unique within the dataset.
        :param str valid_as_of: The date as of which the schema is active.
                               Expected format is YYYY-MM-DD.
        :param dict spark_schema_json: Json schema for the spark dataframe.
        :param list[dict] partitions: Optional. Non empty list of `Partition` as described below.
        :param str description: Optional. Description for the schema.

        .. code-block:: python

                # Partition:
                {
                    "name": "key1",
                    "type": "String"
                }    

        :returns: The registered schema.
        :rtype: collections.namedtuple

        - **Sample**

        .. code-block:: python

                # my_spark_dataframe: some spark dataframe with data being published in the data lake

                my_spark_schema_json = my_spark_dataframe.schema.jsonValue()

                my_schema = client.register_spark_schema(
                    "my-dataset-id",
                    version='1a',
                    valid_as_of='2018-10-31',
                    spark_schema_json=my_spark_schema_json,
                    description="My schema description"
                )
        """

        payload = {
            'version': version,
            'validAsOf': valid_as_of,
            'partitions': partitions,
            'description': description,
        }
        payload.update(spark_schema_json)

        return self._register_schema(dataset_id, payload)

    def get_schema(self, dataset_id, version=None):
        """
        Looks up schema for a dataset by version. In case version is not specified this will fetch schema version having the latest valid_as_of date.
        Throws exception if no schema is registered for the dataset.

        :param str dataset_id: The id of the dataset under which the schema is registered.
        :param str version: Optional. The version of the schema.

        :returns: The schema.
        :rtype: collections.namedtuple

        - **Sample**

        .. code-block:: python

                # Fetch schema version '1a'
                schema = client.get_schema('my_dataset_id', version='1a')

                # Fetch the schema with the latest valid_as_of date
                latest_schema = client.get_schema('my_dataset_id')

        """

        if not dataset_id:
            raise MissingMandatoryArgumentException('dataset_id')

        dataset = self._get_dataset(dataset_id=dataset_id)

        schema_siren = dataset.get_schema(version=version) if version else dataset.get_schema()

        if schema_siren:
            return siren_to_entity(schema_siren)

        raise CatalogueEntityNotFoundException('Schema', {'dataset_id': dataset_id, 'version': version})

    def get_schemas(self, dataset_id, count=100):
        """
        Returns a list of all schemas registered under the dataset. The list is sorted by schema valid_as_of date in descending order.

        :param str dataset_id: The id of the dataset.
        :param int count: Optional count of schemas to be returned. Defaults to 100.

        :returns: List of all schemas registered under the dataset sorted by valid_as_of date in descending order.
        :rtype: list[collections.namedtuple]

        - **Sample**

        .. code-block:: python

                my_schemas = client.get_schemas("my_dataset_id", count=10)
        """

        if not dataset_id:
            raise MissingMandatoryArgumentException('dataset_id')

        ensure_count_is_valid(count)

        dataset = self._get_dataset(dataset_id=dataset_id)

        schemas = dataset.get_schemas(page_size=count).get_entities(rel="schema")

        return [siren_to_entity(s) for s in schemas]

    def delete_schema(self, dataset_id, version):
        """
        Marks a schema version for a dataset as deleted. 

        :param str dataset_id: The id of the dataset under which the schema is registered.
        :param str version: The version of the schema.

        :returns: None

        - **Sample**

        .. code-block:: python

                # Delete schema version '1a'
                client.delete_schema(dataset_id='my_dataset_id', version='1a')

        """

        if not dataset_id:
            raise MissingMandatoryArgumentException('dataset_id')
        if not version:
            raise MissingMandatoryArgumentException('version')

        dataset = self._get_dataset(dataset_id=dataset_id)

        schema = dataset.get_schema(version=version)

        if schema:
            return schema.delete_schema(dataset_id=dataset_id, version=version)

        raise CatalogueEntityNotFoundException('Schema', {'dataset_id': dataset_id, 'version': version})

    def edit_schema(
        self,
        dataset_id,
        version,
        new_version=None,
        valid_as_of=None,
        fields=None,
        type=None,
        partitions=None,
        description=None
    ):
        """
        Updates schema metadata for a dataset.
        If a field is passed as ``None`` then the field will not be updated.

        :param str dataset_id: Id of the dataset for the schema.
        :param str version: Version of the schema being updated.
        :param str new_version: Optional. New version if to be updated. This is a user assigned version name/number. It should be unique within the dataset.
        :param str valid_as_of: Optional. The date as of which the schema is active.
                               Expected format is YYYY-MM-DD.
        :param list[dict] fields: Optional. If provided, a non empty list of `Field` as described below.
        :param str type: Optional. Type of the schema.
        :param list[dict] partitions: Optional. If provided, a non empty list of `Partition` as described below.
        :param str description: Optional. Description for the schema.

        .. code-block:: python

                # Field: Spark compatible
                {
                    "name": "field_a", 			# name of the column.
                    "type": "String", 			# string representation of the type of the column (might be another StructType if it is a nested type).
                    "nullable": True,  			# defaulted to True - A boolean indicating whether the field is nullable or not.
                    "metadata": None			# optional dictionary with metadata for this column.
                }

                # Partition:
                {
                    "name": "key1",
                    "type": "String"
                }    

        :returns: The updated schema.
        :rtype: collections.namedtuple

        - **Sample**

        .. code-block:: python

                # Updating description and valid_as_of date for my schema
                my_updated_schema = client.edit_schema(
                    "my-dataset-id",
                    '1a',
                    valid_as_of='2018-11-05',
                    description="My updated schema description"
                )
        """
        if not dataset_id:
            raise MissingMandatoryArgumentException('dataset_id')
        if not version:
            raise MissingMandatoryArgumentException('version')

        dataset = self._get_dataset(dataset_id=dataset_id)

        schema = dataset.get_schema(version=version)

        if schema:
            schema_as_dict = siren_to_dict(schema)
            payload = {
                'version': new_version,
                'validAsOf': valid_as_of,
                'type': type,
                'fields': fields,
                'partitions': partitions,
                'description': description,
            }

            for key in list(schema_as_dict.keys()):
                if key not in payload:
                    del schema_as_dict[key]

            schema_as_dict.update({k: v for k, v in payload.items() if v is not None})

            result = schema.edit_schema(__json=schema_as_dict)
            return siren_to_entity(result)

        raise CatalogueEntityNotFoundException('Schema', {'dataset_id': dataset_id, 'version': version})



class DatalakeException(Exception):
    pass


class CatalogueEntityNotFoundException(DatalakeException):

    def __init__(self, entity, params=None, message=None):
        if not params and not message:
            message = "%s not found" % entity

        self.message = (
            message or '{} with params `{}` not found'.format(entity, params)
        )

        super(CatalogueEntityNotFoundException, self).__init__(self.message)


class InvalidPayloadException(DatalakeException):
    pass


class S3FileDoesNotExist(DatalakeException):
    def __init__(self, file_path):
        self.message = (
            "Either file at path `%s` does not exist / Potential issue with the bucket policy."
            "Please reach out to Datalake Tech Data Ops user for resolution." % file_path
        )

        super(S3FileDoesNotExist, self).__init__(self.message)


class DownloadDestinationNotValid(DatalakeException):
    """
    Raised when a download destination is not a directory
    """
    pass


class DownloadFailed(DatalakeException):
    pass


class NoAccountSpecified(DatalakeException):

    def __init__(self, accounts):
        self.accounts = accounts
        self.message = (
            "Unable to default the account for access_manager_id, tech_data_ops_id and/or manager_id "
            "due to multiple accounts being attached to this API key. "
            "Your accounts are: %s" % [(a.id, a.name) for a in accounts]
        )
        super(NoAccountSpecified, self).__init__(self.message)


class UnAuthorisedAccessException(DatalakeException):
    pass


class InsufficientPrivilegesException(DatalakeException):
    def __init__(self, message=None):
        self.message = (
            message or 'Insufficient privileges to perform this action'
        )

        super(InsufficientPrivilegesException, self).__init__(self.message)


class MissingMandatoryArgumentException(DatalakeException):
    def __init__(self, argument):
        self.message = "%s is required" % argument
        super(MissingMandatoryArgumentException, self).__init__(self.message)


def is_boto_client_access_denied_error(e):
    # s3fs wraps the boto ClientError under OSError with no particular code. Hence the check
    return isinstance(e, OSError) and 'AccessDenied' in str(e.strerror)



import requests
from requests_toolbelt.adapters import host_header_ssl
from six.moves.urllib.parse import urlparse

from .dli_request_factory_factory import DliRequestFactoryFactory
from pypermedia import HypermediaClient
from pypermedia.client import SirenBuilder
from pypermedia.siren import SirenEntity, _create_action_fn

import jwt
import datetime
import logging


logger = logging.getLogger(__name__)


class Context(object):

    def __init__(self, api_key, api_root, host, auth_key):
        self.api_key = api_key
        self.api_root = api_root
        self.auth_key = auth_key
        self.cache = {}

        self.request_factory = DliRequestFactoryFactory(
            api_root, host, lambda: self.get_header_with_auth()
        ).request_factory

        self.request_session = self.create_request_session(
            api_root,
            host
        )

        self.s3_keys = {}
        self.token_expires_on = self.get_expiration_date(auth_key)

    @staticmethod
    def create_request_session(root, host):
        # build the requests sessions that pypermedia will use
        # to submit requests
        session = requests.Session()

        # when no dns is available and the user is using an ip address
        # to reach the catalogue
        # we need to force the cert validation to be against
        # the host header, and not the host in the uri
        # (we only do this if the scheme of the root is https)
        if host and urlparse(root).scheme == "https":
            session.mount(
                'https://',
                host_header_ssl.HostHeaderSSLAdapter()
            )

        return session

    @staticmethod
    def get_expiration_date(token):
        # use a default_timeout if the token can't be decoded
        # until the proper endpoint is added on the catalog
        default_timeout = (
            datetime.datetime.utcnow() +
            datetime.timedelta(minutes=55)
        )

        try:
            # get the expiration from the jwt auth token
            decoded_token = jwt.get_unverified_header(token)
            if 'exp' not in decoded_token:
                return default_timeout 

            return datetime.datetime.fromtimestamp(
                decoded_token['exp']
            )

        except Exception:
            return default_timeout

    @property
    def session_expired(self):
        # by default we don't want to fail if we could not decode the token
        # so if the ``token_expiry`` is undefined we assume the session
        # is valid
        if not self.token_expires_on:
            return False

        return datetime.datetime.utcnow() > self.token_expires_on

    def get_header_with_auth(self):
        auth_header = "Bearer {}".format(self.auth_key)
        return {"Authorization": auth_header}

    def uri_with_root(self, relative_path):
        return "{}/{}".format(self.api_root, relative_path)

    # not a decorator since we want to access the local
    # cache instance (which needs to be evicted when the session/context
    # is recreated)
    def memoized(self, func, key=None):
        key = key or func.__name__
        if key not in self.cache:
            self.cache[key] = func()

        return self.cache[key]

    def get_root_siren(self):
        return HypermediaClient.connect(
            self.api_root,
            session=self.request_session,
            request_factory=self.request_factory,
            builder=PatchedSirenBuilder
        )


class PatchedSirenBuilder(SirenBuilder):

    def _construct_entity(self, entity_dict):
        """
        We need to patch the actions as there is no ``radio`` support
        on the current pypermedia version.

        To avoid code duplication, this function will attempt to call the parent
        and replace the created actions with our custom ones.
        """

        # pypermedia does not like any custom attributes
        # not even those in the spec
        for action in entity_dict.get("actions", []):
            if "allowed" in action:
                del action["allowed"]

        entity = super(PatchedSirenBuilder, self)._construct_entity(entity_dict)

        return PatchedSirenEntity(
            classnames=entity.classnames,
            properties=entity.properties,
            actions=entity.actions,
            links=entity.links,
            entities=entity.entities,
            rel=entity.rel,
            verify=entity.verify,
            request_factory=entity.request_factory
        )


class PatchedSirenEntity(SirenEntity):

    def as_python_object(self):
        ModelClass = type(str(self.get_primary_classname()), (), self.properties)

        siren_builder = PatchedSirenBuilder(verify=self.verify, request_factory=self.request_factory)
        # add actions as methods
        for action in self.actions:
            method_name = SirenEntity._create_python_method_name(action.name)
            method_def = _create_action_fn(action, siren_builder)
            setattr(ModelClass, method_name, method_def)

        # add links as methods
        for link in self.links:
            for rel in link.rel:
                method_name = SirenEntity._create_python_method_name(rel)
                siren_builder = PatchedSirenBuilder(verify=self.verify, request_factory=self.request_factory)
                method_def = _create_action_fn(link, siren_builder)

                setattr(ModelClass, method_name, method_def)

        def get_entity(obj, rel):
            matching_entities = self.get_entities(rel) or []
            for x in matching_entities:
                yield x.as_python_object()

        setattr(ModelClass, 'get_entities', get_entity)

        return ModelClass()



import logging
import uuid
from functools import partial

from dli.client.exceptions import (
    CatalogueEntityNotFoundException,
    MissingMandatoryArgumentException,
    NoAccountSpecified,
)
from tests.common import SdkIntegrationTestCase

logger = logging.getLogger(__name__)


class CollectionFunctionsTestCase(SdkIntegrationTestCase):

    def setUp(self):
        super(CollectionFunctionsTestCase, self).setUp()
        self.create = partial(
            self.client.create_collection,
            name="CollectionFunctionsTestCase" + str(uuid.uuid4()),
            description="My collection description",
            manager_id="datalake-mgmt",
            documentation='This collection should be documented!',
            keywords=['test', 'abc']
        )

    def test_can_create_collection(self):
        collection = self.create()
        self.assertIsNotNone(collection)

    def test_cannot_create_collection_if_api_key_has_multiple_accounts(self):
        with self.assertRaises(NoAccountSpecified):
            self.create(manager_id=None)

    def test_edit_collection_should_unset_field_when_passed_none(self):
        collection = self.create(documentation='I have some docs')
        updated_collection = self.client.edit_collection(collection.collection_id, documentation=None)
        self.assertTrue('documentation' not in updated_collection._fields)

    def test_edit_collection_should_ignore_extra_keys_and_succeed_when_unknown_key_passed_with_kwargs(self):
        collection = self.create()
        updated_collection = self.client.edit_collection(collection.collection_id, i_do_not_exist='sabotage')
        self.assertTrue('i_do_not_exist' not in updated_collection._fields)

    def test_can_get_collection_details_by_id_or_name(self):
        collection = self.create()
        collection_by_id = self.client.get_collection(collection.collection_id)
        collection_by_id_again = self.client.get_collection(id=collection.collection_id)
        collection_by_name = self.client.get_collection(name=collection.name)

        self.assertTrue(all(c is not None for c in [collection_by_id, collection_by_id_again, collection_by_name]))
        self.assertEqual(collection_by_id.collection_id, collection_by_name.collection_id)
        self.assertEqual(collection_by_id.collection_id, collection_by_id_again.collection_id)

    def test_cannot_get_unknown_collection(self):
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.get_collection('unknown')

    def test_cannot_get_collection_without_collection_id_or_name(self):
        with self.assertRaises(ValueError):
            self.client.get_collection(None)
            self.client.get_collection('')
            self.client.get_collection(id='', name='')

    def test_cannot_edit_unknown_collection(self):
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.edit_collection('unknown')

    def test_cannot_edit_collection_without_collection_id(self):
        with self.assertRaises(MissingMandatoryArgumentException):
            self.client.edit_collection(None)

    def test_can_edit_collection(self):
        collection = self.create()
        updated_collection = self.client.edit_collection(collection.collection_id,
                                                         description='Updated collection description')

        self.assertEqual(updated_collection.description, 'Updated collection description')
        self.assertEqual(collection.collection_id, updated_collection.collection_id)
        self.assertEqual(collection.name, updated_collection.name)
        # Collection manager still the same
        self.assertEqual(collection.manager_id, updated_collection.manager_id)

    def test_can_edit_collection_manager(self):
        collection = self.create()
        updated_collection = self.client.edit_collection(collection.collection_id, manager_id='iboxx')

        self.assertEqual(collection.collection_id, updated_collection.collection_id)
        self.assertEqual(updated_collection.manager_id, 'iboxx')

    def test_cannot_delete_unknown_collection(self):
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.delete_collection('unknown')

    def test_cannot_delete_collection_without_collection_id(self):
        with self.assertRaises(MissingMandatoryArgumentException):
            self.client.delete_collection(None)

    def test_can_delete_collection(self):
        collection = self.create()
        self.client.delete_collection(collection.collection_id)

        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.get_collection(collection.collection_id)

    def test_cannot_get_packages_for_collection_with_invalid_page_count(self):
        self.assert_page_count_is_valid_for_paginated_resource_actions(
            lambda c: self.client.get_packages_for_collection(collection_id='some_collection', count=c))

    def test_cannot_get_packages_for_unknown_collection(self):
        with self.assertRaises(CatalogueEntityNotFoundException):
            self.client.get_packages_for_collection('unknown')

    def test_cannot_get_packages_for_collection_without_collection_id(self):
        with self.assertRaises(MissingMandatoryArgumentException):
            self.client.get_packages_for_collection(None)

    def test_get_packages_for_empty_collection_returns_empty_list(self):
        my_empty_collection = self.create()
        self.assertEqual(self.client.get_packages_for_collection(my_empty_collection.collection_id), [])

    def test_can_get_packages_for_collection(self):
        collection_of_my_packages = self.create()
        package_one = self.create_package('Package One', collection_ids=[collection_of_my_packages.collection_id])
        package_two = self.create_package('Package Two', collection_ids=[collection_of_my_packages.collection_id])

        my_packages = self.client.get_packages_for_collection(collection_of_my_packages.collection_id)
        self.assertEqual(len(my_packages), 2)

        my_package_ids = [p.package_id for p in my_packages]
        self.assertIn(package_one, my_package_ids)
        self.assertIn(package_two, my_package_ids)



import logging
from dli.siren import siren_to_entity, siren_to_dict
from dli.client.exceptions import (
    CatalogueEntityNotFoundException,
    MissingMandatoryArgumentException
)
from dli.client.utils import ensure_count_is_valid, filter_out_unknown_keys, to_camel_cased_dict

logger = logging.getLogger(__name__)


class PackageFunctions:

    _KNOWN_FIELDS = {"name",
                     "description",
                     "keywords",
                     "topic",
                     "access",
                     "internalData",
                     "dataSensitivity",
                     "contractIds",
                     "termsAndConditions",
                     "derivedDataNotes",
                     "derivedDataRights",
                     "distributionNotes",
                     "distributionRights",
                     "internalUsageNotes",
                     "internalUsageRights",
                     "documentation",
                     "publisher",
                     "techDataOpsId",
                     "accessManagerId",
                     "managerId",
                     "collectionIds",
                     "intendedPurpose"}
    """
    A mixin providing common package operations
    """

    @property
    def __root(self):
        return self.ctx.memoized(
            self.get_root_siren().packages_root,
            "packages_root"
        )

    def get_package(self, id=None, name=None):
        """
        Fetches package metadata for an existing package.

        :param str id: The id of the package.
        :param str name: The name of the package.

        :returns: A package instance
        :rtype: collections.namedtuple

        - **Sample**

        .. code-block:: python

                # Look up by package id
                package = client.get_package('my_package_id') 
                # or
                package = client.get_package(id='my_package_id')

                # Alternatively look up by package name
                package = client.get_package(name='my_package')

        """
        if id is not None:
            return siren_to_entity(self._get_package(package_id=id))

        if name is not None:
            return siren_to_entity(self._get_package(name=name))

        raise ValueError("Either package id or name must be specified to look up package")

    def get_package_datasets(self, package_id, count=100):
        """
        Returns a list of all datasets registered under a package.

        :param str package_id: The id of the package.
        :param int count: Optional. Count of datasets to be returned. Defaults to 100.

        :returns: List of all datasets registered under the package.
        :rtype: list[collections.namedtuple]

        - **Sample**

        .. code-block:: python

                datasets = client.get_package_datasets(
                    package_id,
                    count=100
                )
        """
        if not package_id:
            raise MissingMandatoryArgumentException('package_id')

        ensure_count_is_valid(count)

        package = self._get_package(package_id=package_id)

        datasets = package.get_datasets(
            page_size=count
        ).get_entities(rel="dataset")
        return [siren_to_entity(d) for d in datasets]

    def register_package(
        self,
        name,
        description,
        topic,
        access,
        internal_data,
        data_sensitivity,
        terms_and_conditions,
        publisher,
        access_manager_id,
        tech_data_ops_id,
        manager_id,
        keywords=None,
        contract_ids=None,
        derived_data_notes=None,
        derived_data_rights=None,
        distribution_notes=None,
        distribution_rights=None,
        internal_usage_notes=None,
        internal_usage_rights=None,
        documentation=None,
        collection_ids=None,
        intended_purpose=None
    ):
        """
        Submit a request to create a new package in the Data Catalogue.

        Packages are parent structures that contain metadata relating
        to a group of Datasets.

        See description for each parameter, and whether they are optional or mandatory.

        :param str name: A descriptive name of a package. It should be unique across the Data Catalogue.
        :param str description: A short description of a package.
        :param str topic: Topic the data in the package is about. Not applicable, if the package is not industry specific.
        :param str access: Accepted values are: `Restricted` or `Unrestricted`.
                            If access to the package is flagged as `Restricted`,
                            access manager_id will have to grant or deny access to the underlying data.
                            If access is flagged as `Unrestricted`, user will be able to gain
                            access instantaneously after submitting the access request form.
        :param str internal_data: Accepted values are: `Yes`, `No` or `Both`.
                            Package is marked as `Yes` if underlying data is created internally at IHS Markit, 
                            or `No` if externally, e.g. S&P, Russell, etc.
        :param str data_sensitivity: Accepted values are: `Private`, `Public` or `Top Secret`. Sensitivity level of the data contained within the package
        :param str terms_and_conditions: Not provided, empty or blank will result in ValueError.
                            Default T&C can be obtained through `get_default_package_terms_and_conditions` function.
        :param str publisher: Business unit or legal entity responsible for the content.
                              For example, S&P, Dow Jones, IHS Markit.
        :param str access_manager_id: Account ID for the Data Lake Account representing
                                      IHS Markit business unit that is responsible for managing access to the packages on Data Catalogue.
        :param str tech_data_ops_id: Account ID for the Data Lake Account representing
                                IHS Markit business unit that is responsible for uploading the data to Data Lake.
        :param str manager_id: Mandatory Account ID for the Data Lake Account representing
                            IHS Markit business unit that is responsible for creating and maintaining metadata for packages and datasets
                            on Data Catalogue.
        :param list[str] keywords: Optional. List of keywords that can be used to find this
                         package through the search interface.

        :param list[str] contract_ids: Optional. Internally, this will be the Salesforce contract ID and/or CARM ID. Externally, this could be any ID.
        :param str derived_data_notes: Optional. Provides details, comments on derived data.
                                   Extension to the Derived Data Rights field.
        :param str derived_data_rights: Optional. Accepted values are `Yes`, `No`, `With Limitations`, `N/A`.
                                    A flag to indicate whether we have rights to derived data.
        :param str distribution_notes: Optional. Provides details, comments on data distribution rights.
                                   Extension to the Distribution Rights field.
        :param str distribution_rights: Optional. Accepted values are `Yes`, `No`, `With Limitations`, `N/A`.
                                    A flag to indicate whether data can be distributed.
        :param str internal_usage_notes: Optional. Provides details, comments on internal data usage.
                                     Extension to Internal Usage Rights.
        :param str internal_usage_rights: Optional. Accepted values are: `Yes`, `No`, `With Limitations`, `N/A`.
                                      A flag to indicate whether data can be used internally.
        :param str documentation: Optional. Documentation about this package in markdown format.
        :param list[str] collection_ids: Optional. List of ids of collections attached to this package.
        :param str intended_purpose: Optional. Provides details about intended usage of the data contained 
                                     in the package, e.g. permanent storage, temporary storage, POC.
        :returns: Created package
        :rtype: collections.namedtuple

        - **Sample**

        .. code-block:: python

                package = client.register_package(
                    name="my package",
                    description="my package description",
                    topic="Automotive",
                    access="Restricted",
                    internal_data="Yes",
                    data_sensitivity="Public",
                    terms_and_conditions="Terms",
                    publisher="my publisher"
                )
        """

        self._validate_terms_and_conditions(terms_and_conditions)

        payload = {
            "name": name,
            "description": description,
            "keywords": keywords,
            "topic": topic,
            "access": access,
            "internalData": internal_data,
            "dataSensitivity": data_sensitivity,
            "contractIds": contract_ids,
            "termsAndConditions": terms_and_conditions,
            "derivedDataNotes": derived_data_notes,
            "derivedDataRights": derived_data_rights,
            "distributionNotes": distribution_notes,
            "distributionRights": distribution_rights,
            "internalUsageNotes": internal_usage_notes,
            "internalUsageRights": internal_usage_rights,
            "documentation": documentation,
            "publisher": publisher,
            "techDataOpsId": tech_data_ops_id,
            "accessManagerId": access_manager_id,
            "managerId": manager_id,
            "collectionIds": collection_ids,
            "intendedPurpose": intended_purpose
        }

        payload = {k: v for k, v in payload.items() if v is not None}
        return siren_to_entity(self.__root.create_package(__json=payload))

    def edit_package(
        self,
        package_id,
        **kwargs
    ):
        """
        Updates one or more fields in a package.
        If keyword argument is not specified field keeps its old value.
        Optional enum and text type fields can be unset by passing ``None``.

        :param str package_id: Package ID of the package being edited.

        Keyword arguments:

        :keyword str name: Optional. A descriptive name of a package. It should be unique across the Data Catalogue.
        :keyword str description: Optional. A short description of a package.
        :keyword str topic: Optional. Topic the data in the package is about. Not applicable, if the package is not industry specific.
        :keyword str access: Optional. Accepted values are: `Restricted` or `Unrestricted`.
                            If access to the package is flagged as `Restricted`,
                            access manager will have to grant or deny access to the underlying data.
                            If access is flagged as `Unrestricted`, user will be able to gain
                            access instantaneously after submitting the access request form.
        :keyword str internal_data: Optional. Accepted values are: `Yes`, `No` or `Both`.
                            Package is marked as `Yes` if underlying data is created internally at IHS Markit, 
                            or `No` if externally, e.g. S&P, Russell, etc.
        :keyword str data_sensitivity: Optional. Accepted values are: `Private`, `Public` or `Top Secret`. Sensitivity level of the data contained within the package
        :keyword str terms_and_conditions: Optional. Terms and conditions for the package. Default T&C can be obtained through `get_default_package_terms_and_conditions`
                                        function. None, empty or blank string will result in ValueError.
        :keyword str publisher: Optional. Business unit or legal entity responsible for the content.
                              For example, S&P, Dow Jones, IHS Markit.        
        :keyword list[str] keywords: Optional. List of keywords that can be used to find this
                         package through the search interface.
        :keyword str access_manager_id: Optional. Account ID for the Data Lake Account representing IHS Markit
                        business unit that is responsible for managing access
                        to the packages on Data Catalogue.
        :keyword str tech_data_ops_id: Optional. Account ID for the Data Lake Account representing
                          IHS Markit business unit that is responsible for uploading
                          the data to Data Lake.
        :keyword str manager_id: Optional. Account ID for the Data Lake Account representing IHS Markit
                        business unit that is responsible for creating and
                        maintaining metadata for packages and datasets on Data Catalogue.
        :keyword list[str] contract_ids: Optional. Internally, this will be the Salesforce contract ID and/or CARM ID. Externally, this could be any ID.
        :keyword str derived_data_notes: Optional. Provides details, comments on derived data.
                                   Extension to the Derived Data Rights field.
        :keyword str derived_data_rights: Optional. Accepted values are `Yes`, `No`, `With Limitations`, `N/A`.
                                    A flag to indicate whether we have rights to derived data.
        :keyword str distribution_notes: Optional. Provides details, comments on data distribution rights.
                                   Extension to the Distribution Rights field.
        :keyword str distribution_rights: Optional. Accepted values are `Yes`, `No`, `With Limitations`, `N/A`.
                                    A flag to indicate whether data can be distributed.
        :keyword str internal_usage_notes: Optional. Provides details, comments on internal data usage.
                                     Extension to Internal Usage Rights.
        :keyword str internal_usage_rights: Optional. Accepted values are: `Yes`, `No`, `With Limitations`, `N/A`.
                                      A flag to indicate whether data can be used internally.
        :keyword str documentation: Optional. Documentation about this package in markdown format.
        :keyword list[str] collection_ids: Optional. List of ids of collections attached to this package.
        :keyword str intended_purpose: Optional. Provides details about intended usage of the data contained
                                     in the package, e.g. permanent storage, temporary storage, POC.

        :returns: The updated Package.
        :rtype: collections.namedtuple

        - **Sample**

        .. code-block:: python

                package = client.edit_package(
                    package_id="my-package-id",
                    description="Updated my package description",
                )
        """
        if not package_id:
            raise MissingMandatoryArgumentException('package_id')

        fields = filter_out_unknown_keys(to_camel_cased_dict(kwargs), PackageFunctions._KNOWN_FIELDS)
        self._validate_fields(fields)

        package = self._get_package(package_id=package_id)
        # we can't just post back the siren object for some reason
        # as it can't be deserialised
        package_as_dict = siren_to_dict(package)

        # clean the package dict with fields that aren't known to us
        for key in list(package_as_dict.keys()):
            if key not in PackageFunctions._KNOWN_FIELDS:
                del package_as_dict[key]

        package_as_dict.update(fields)

        result = package.edit_package(__json=package_as_dict)
        return siren_to_entity(result)

    def delete_package(self, package_id):
        """
        Performs deletion of an existing package. This will delete all underlying datasets for the package as well.

        :param str package_id: The id of the package to be deleted.

        :returns: None

        - **Sample**

        .. code-block:: python

                client.delete_package(package_id)

        """
        if not package_id:
            raise MissingMandatoryArgumentException('package_id')

        package = self._get_package(package_id=package_id)
        if package:
            package.delete_package(package_id=package_id)

    @staticmethod
    def get_default_package_terms_and_conditions():
        """
        Returns a string representing the default Terms And Conditions for packages created in DataLake.

        :returns: The default DataLake Terms And Conditions
        :rtype: str
        """
        return ('By submitting this Data request and checking the "Accept Terms and Conditions" '
            'box, you acknowledge and agree to the following:\n'
            '\n'
            '* To promptly notify the relevant Access Manager/Producer of your intended use '
            'of the Data;\n'
            '* To obtain the terms and conditions relevant to such use for such Data from '
            'the Producer;\n'
            '* To distribute such terms and conditions to each member of your '
            'Consumer Group who may use the Data;\n'
            '* To use the Data solely for such intended use, subject to such terms and '
            'conditions;\n'
            '* To ensure that the Data is only accessed by members of your Consumer Group, '
            'and only used by such members for such intended use, subject to such terms and '
            'conditions;\n'
            '* To adhere to any additional requests of Producer with respect to the Data '
            '(including but not limited to ceasing use of the Data and deleting the Data, '
            'and ensuring other members of the Consumer Group do so, upon revocation of your '
            'license by Producer).\n'
            '\n'
            'Please refer to the <a href="/terms-of-use" target="_blank">EULA</a> for any '
            'defined terms used above. '
            'The <a href="/terms-of-use" target="_blank">EULA</a> '
            'is the document you agreed to adhere to by accessing the Lake.')

    #
    # Private functions
    #
    def _get_package(self, **kwargs):

        package = self.__root.get_package(**kwargs)

        if not package:
            raise CatalogueEntityNotFoundException('Package', params=kwargs)

        return package

    @staticmethod
    def _validate_fields(fields):
        if 'termsAndConditions' in fields:
            PackageFunctions._validate_terms_and_conditions(fields.get("termsAndConditions"))

    @staticmethod
    def _validate_terms_and_conditions(field):
        if not field or not field.strip():
            raise ValueError("Terms and conditions must be defined "
                             "and be non empty, non blank string")



import unittest
import mock
from requests import Session

from dli.client import dli_client


class TestAuth(unittest.TestCase):

    @mock.patch.object(Session, 'post')
    def test_handle_auth_error(self, post):
        text = 'Some server error message'
        post.return_value = mock.Mock(status_code=401, text=text)
        with self.assertRaises(dli_client.AuthenticationFailure) as cm:
            dli_client._get_auth_key('api-key', 'api-root')

        self.assertEqual(str(cm.exception), text)

    @mock.patch.object(Session, 'post')
    def test_handle_auth_error_for_403_response(self, post):
        """
        Should show a generic response message
        """
        post.return_value = mock.Mock(status_code=403)
        with self.assertRaises(dli_client.AuthenticationFailure) as cm:
            dli_client._get_auth_key('api-key', 'api-root')
            
        exception_message = str(cm.exception)
        self.assertIn(dli_client.AuthenticationFailure.GENERIC_ERROR_MESSAGE, exception_message)
        self.assertIn('api-root', exception_message)

    @mock.patch.object(Session, 'post')
    def test_auth_error_for_unknown_auth_error(self, post):
        """
        Should show a generic response message if no error message
        """
        post.return_value = mock.Mock(status_code=401, text='')

        with self.assertRaises(dli_client.AuthenticationFailure) as cm:
            dli_client._get_auth_key('api-key', 'api-root')

        self.assertEqual(
            str(cm.exception),
            dli_client.AuthenticationFailure.GENERIC_ERROR_MESSAGE
        )


from dli.client.entities_mixin import EntityMixin
from dli.client.search_functions import SearchFunctions
from dli.client.me_functions import MeFunctions


"""
Define the client mixin here which consists of the mixin aggregated by 
function groups, for e.g entities, search, user etc.
"""
class ClientMixin(EntityMixin, MeFunctions, SearchFunctions):
    pass



class DatasetBuilder:
    """
    Dataset is grouping of related datafiles. It provides user with metadata required to consume and use the data.

    This builder object sets sensible defaults and exposes
    helper methods on how to configure its storage options..

    :param str package_id: Package ID to which the dataset belongs to.
    :param str name: A descriptive name of the dataset. It should be unique within the package the dataset
                    belongs to.
    :param str description: A short description of the dataset.
    :param str content_type: A way for the data steward to classify the type of data in the dataset
                        (e.g. pricing).
    :param str data_format: The format of the data: `CSV`, `IMAGE`, `JSON`, `PARQUET`, `TXT`, `XML`, `Other`.
    :param str publishing_frequency: The internal on which data is published. Possible values are: `Hourly`,
                                        `Daily`, `Weekday`, `Weekly`, `Monthly`, `Quarterly`, `Yearly`, `Not Specified`.
    :param list[str] taxonomy: A list of segments to be used for a taxonomy,
                                    the Data-<< Organization >>-<< topic >> prefix will be applied by the catalogue  For
                                    a taxonomy of Data-IHS Markit-Financial Markets-Credit-CDS, you would provide
                                    `taxonomy=["Credit", "CDS"]`
    :param list[str] keywords: Optional. User-defined list of separated keywords that can be used to find this dataset
                            through the search interface.
    :param str naming_convention: Optional. Key for how to read the dataset name.
    :param str documentation: Optional. Documentation about this dataset in markdown format.
    :param str load_type: Optional. Whether each datafile in this dataset should be considered as a full version of a
                        dataset or a delta or increment. Accepted types are `Full Load`, `Incremental Load`
    """

    def __init__(
        self,
        package_id,
        name,
        description,
        content_type,
        data_format,
        publishing_frequency,
        taxonomy,
        keywords=None,
        naming_convention=None,
        documentation=None,
        load_type=None
    ):
        self.payload = {
            "packageId": package_id,
            "name": name,
            "description": description,
            "keywords": keywords,
            "contentType": content_type,
            "location": None,
            "dataFormat": data_format,
            "publishingFrequency": publishing_frequency,
            "namingConvention": naming_convention,
            "documentation": documentation,
            "taxonomy": taxonomy,
            "loadType": load_type
        }

    def with_external_s3_storage(
        self,
        bucket_name,
        aws_account_id,
        prefix
    ):
        """
        Indicate that the dataset will be stored
        in a self-managed S3 bucket outside of the Data Lake.

        :param str bucket_name: Name of the bucket you want to link to this dataset.
        :param str aws_account_id: The AWS account id where this bucket currently resides.
                                   This account needs to be registed on the data lake previously
                                   and your account should have permissions to use it.
        :param str prefix: A valid path that specifies the absolute parent
                           for files in this dataset.
                           This value will be used when issuing access tokens so 
                           it is essential that it is as constrained as possible.
                           Cannot end with slash ("/").

        :rtype: dli.client.builders.DatasetBuilder

        - **Sample**

        .. code-block:: python

                from dli.client.builders import DatasetBuilder

                builder = DatasetBuilder(
                                package_id="package-id",
                                name="my test dataset",
                                description="My dataset description",
                                content_type="Pricing",
                                data_format="CSV",
                                publishing_frequency="Weekly",
                                taxonomy=["Credit", "CDS"]
                        )
                builder = builder.with_external_s3_storage(
                    bucket_name="external-s3-bucket-name",
                    aws_account_id=123456789,
                    prefix="/economic-data-package/my-test-dataset"
                )
                dataset = client.register_dataset(builder)
        """
        location_builder = DatasetLocationBuilder().with_external_s3_storage(
            bucket_name=bucket_name,
            aws_account_id=aws_account_id,
            prefix=prefix
        )
        self.payload.update(location_builder.build())
        return self

    def with_external_storage(self, location):
        """
        Allows specifying a non S3 location where
        the dataset resides.

        The location will be kept for informational purposes only.

        :param str location: A connection string or identifier where the dataset resides.

        :rtype: dli.client.builders.DatasetBuilder

        - **Sample**

        .. code-block:: python

                from dli.client.builders import DatasetBuilder

                builder = DatasetBuilder(
                                package_id="package-id",
                                name="my test dataset",
                                description="My dataset description",
                                content_type="Pricing",
                                data_format="CSV",
                                publishing_frequency="Weekly",
                                taxonomy=["Credit", "CDS"]
                        )
                builder = builder.with_external_storage("external-storage-location")
                dataset = client.register_dataset(builder)
        """
        location_builder = DatasetLocationBuilder().with_external_storage(location)
        self.payload.update(location_builder.build())
        return self

    def build(self):
        _validate_location_exists_in_payload(self.payload)
        # clean not set entries
        payload = {k: v for k, v in self.payload.items() if v is not None}
        return payload


class DatasetLocationBuilder:
    """
        A simple builder to specify dataset location.
    """

    def __init__(self):
        self.payload = {}

    def with_external_s3_storage(
        self,
        bucket_name,
        aws_account_id,
        prefix
    ):
        """
        Indicate that the dataset will be stored
        in a self-managed S3 bucket outside of the Data Lake.

        :param str bucket_name: Name of the bucket you want to link to this dataset.
        :param str aws_account_id: The AWS account id where this bucket currently resides.
                                   This account needs to be registed on the data lake previously
                                   and your account should have permissions to use it.
        :param str prefix: A vaild path that specifies the absolute parent
                           for files in this dataset.
                           This value will be used when issuing access tokens so 
                           it is essential that it is as constrained as possible.
                           Cannot end with slash ("/").

        :rtype: dli.client.builders.DatasetLocationBuilder

        - **Sample**

        .. code-block:: python

                from dli.client.builders import DatasetLocationBuilder

                location_builder = DatasetLocationBuilder().with_external_s3_storage(
                        bucket_name="external-s3-bucket-name",
                        aws_account_id=123456789,
                        prefix="/economic-data-package/my-test-dataset"
                    )
                # Build the location object
                location = location_builder.build()
        """
        self.payload["location"] = {
            "type": "S3",
            "owner": {
                "awsAccountId": str(aws_account_id)
            },
            "bucket": bucket_name,
            "prefix": prefix
        }
        return self

    def with_external_storage(self, location):
        """
        Allows specifying a non S3 location where
        the dataset resides.

        The location will be kept for informational purposes only.

        :param str location: A connection string or identifier where the dataset resides.

        :rtype: dli.client.builders.DatasetLocationBuilder

        - **Sample**

        .. code-block:: python

                from dli.client.builders import DatasetLocationBuilder

                location_builder = DatasetLocationBuilder().with_external_storage("external-storage-location")
                # Build the location object
                location = location_builder.build()
        """
        self.payload["location"] = {
            "type": "Other",
            "source": location
        }
        return self

    def build(self):
        _validate_location_exists_in_payload(self.payload)
        # clean not set entries
        payload = {k: v for k, v in self.payload.items() if v is not None}
        return payload


def _validate_location_exists_in_payload(payload):
    if not payload.get("location"):
        raise Exception(
            "No storage option was specified. Please use one of the following methods: "
            "with_external_s3_storage`, `with_external_storage`"
        )



try:
    from pathlib import Path
except ImportError:
    from pathlib2 import Path  # python 2 backport


def makedirs(path, exist_ok=False):
    Path(path).mkdir(parents=True, exist_ok=exist_ok)


def path_for(*parts):
    return Path(*parts)


def to_camel_case(snake_str):
    split = snake_str.split('_')
    camel_case_parts = [split[0].lower()] + [s.title() for s in split[1:]]
    return ''.join(camel_case_parts)


def to_camel_cased_dict(dictionary):
    return {to_camel_case(key): value for (key, value) in dictionary.items()}


def filter_out_unknown_keys(dictionary, keys_set):
    return {k: v for (k, v) in dictionary.items() if k in keys_set}

def ensure_count_is_valid(count):
    count = int(count)
    if count <= 0:
        raise ValueError("`count` should be a positive integer")



import logging
import os
import six

from .common import SdkIntegrationTestCase


class MeFunctionsTestCase(SdkIntegrationTestCase):
    
    def assert_my_entities_are_returned(self, func):
        self.assertGreater(len(func()), 0)
        self.assertEquals(len(func(count=1)), 1)

    def test_get_my_consumed_packages_validates_page_size(self):
        self.assert_page_count_is_valid_for_paginated_resource_actions(lambda c: self.client.get_my_consumed_packages(count=c))

    def test_my_consumed_packages_returns_my_consumed_packages(self):
        package_id = self.create_package('test_consumed_packages', access="Unrestricted")
        package = self.client._get_package(package_id=package_id)
        package.request_access(
            accountId='iboxx',
            operation='request',
            comment='Hello there!'
        )
        self.assert_my_entities_are_returned(self.client.get_my_consumed_packages)

    def test_get_my_packages_validates_page_size(self):
        self.assert_page_count_is_valid_for_paginated_resource_actions(lambda c: self.client.get_my_packages(count=c))

    def test_get_my_packages_returns_packages(self):
        self.create_package("test_me_functions")
        self.assert_my_entities_are_returned(self.client.get_my_packages)

    def test_get_my_collections_validates_page_size(self):
        self.assert_page_count_is_valid_for_paginated_resource_actions(lambda c: self.client.get_my_collections(count=c))

    def test_get_my_collections_returns_collections(self):
        self.create_collection("test_get_my_collection_returns_collections")
        self.assert_my_entities_are_returned(self.client.get_my_collections)


import logging

from dli.client.exceptions import (
    CatalogueEntityNotFoundException,
    MissingMandatoryArgumentException,
    NoAccountSpecified,
)
from dli.siren import siren_to_entity, siren_to_dict

from dli.client.utils import ensure_count_is_valid, filter_out_unknown_keys, to_camel_cased_dict


logger = logging.getLogger(__name__)


class CollectionFunctions:
    _KNOWN_FIELDS = set(["name", "description", "managerId", "documentation", "keywords"])
    @property
    def __root(self):
        return self.ctx.memoized(
            self.get_root_siren().collections_root,
            "collections_root"
        )

    def create_collection(
        self,
        name,
        description,
        manager_id=None,
        documentation=None,
        keywords=None
    ):
        """
        Submit a request to create a new collection in the Data Catalogue.
        
        A Collection defines a group of packages that share thematic information. 
        For example, data for a specific asset class (i.e. CDS) could be a collection.

        See description for each parameter, and whether they are optional or mandatory.

        :param str name: A descriptive name of the collection. It should be unique across the Data Catalogue.
        :param str description: A short description of the collection.
        :keyword str manager_id: Optional. Defaults to your Data Lake Account if none provided. Account ID for the Data Lake Account representing
                            IHS Markit business unit that is responsible for creating and maintaining metadata for the collection in the Data Catalogue.
        :keyword str documentation: Optional. Documentation about the collection in markdown format.
        :keyword list[str] keywords: Optional. List of user-defined terms that can be used to find this
                         collection through the search interface.

        :returns: a newly created Collection
        :rtype: collections.namedtuple

        - **Sample**

        .. code-block:: python

                collection = client.create_collection(
                    name="my collection",
                    description="my collection description",
                )
        """

        if not manager_id:
            accounts = self.get_my_accounts()
            if len(accounts) > 1:
                raise NoAccountSpecified(accounts)

            manager_id = accounts[0].id
        
        payload = {
            "name": name,
            "description": description,
            "managerId": manager_id,
            "documentation": documentation,
            "keywords": keywords,           
        }

        payload = {k: v for k, v in payload.items() if v is not None}
        return siren_to_entity(self.__root.create_collection(__json=payload))

    def get_collection(self, id=None, name=None):
        """
        Fetches collection details for an existing collection. 

        :param str id: The id of the collection.
        :param str name: The name of the collection.

        :returns: NamedTuple representing a Collection instance
        :rtype: collections.namedtuple

        - **Sample**

        .. code-block:: python

                # Look up by collection id
                collection = client.get_collection('my_collection_id') 
                # or
                collection = client.get_collection(id='my_collection_id')

                # Alternatively look up by collection name
                collection = client.get_collection(name='my_collection')
                
        """
        if id:
            return siren_to_entity(self._get_collection(collection_id=id))
        
        if name:
            return siren_to_entity(self._get_collection(name=name))
        
        raise ValueError("Either collection id or name must be specified and must be non-empty in order to look up collection")

    def _get_collection(self, **kwargs):

        collection = self.__root.get_collection(**kwargs)

        if not collection:
            raise CatalogueEntityNotFoundException('Collection', params=kwargs)

        return collection

    def edit_collection(
        self,
        collection_id,
        **kwargs
    ):
        """
        Updates one or more fields in a collection.
        If keyword argument is not specified field keeps its old value.
        Optional enum and text type fields can be unset by passing ``None``.

        :param str collection_id: ID of the collection being edited.

        Keyword arguments:

        :param str name: Optional. A descriptive name of the collection. It should be unique across the Data Catalogue.
        :param str description: Optional. A short description of the collection.
        :param str manager_id: Optional. Account ID for the Data Lake Account representing IHS Markit business unit that is responsible
                            for creating and maintaining metadata for the collection in the Data Catalogue.
        :param str documentation: Optional. Documentation about the collection in markdown format.
        :param list[str] keywords: Optional. List of user-defined terms that can be used to find this
                            collection through the search interface.

        :returns: Updated collection
        :rtype: collections.namedtuple

        - **Sample**

        .. code-block:: python

                collection = client.edit_collection(
                    collection_id="my-collection-id",
                    description="Updated my collection description",
                )
        """
        if not collection_id:
            raise MissingMandatoryArgumentException('collection_id')

        collection = self._get_collection(collection_id=collection_id)

        fields = filter_out_unknown_keys(to_camel_cased_dict(kwargs), CollectionFunctions._KNOWN_FIELDS)
        collection_as_dict = siren_to_dict(collection)

        # drop the fields from collection dict that are not being edited
        for key in list(collection_as_dict.keys()):
            if key not in CollectionFunctions._KNOWN_FIELDS:
                del collection_as_dict[key]

        collection_as_dict.update(fields)

        result = collection.edit_collection(__json=collection_as_dict)
        return siren_to_entity(result)

    def delete_collection(self, collection_id):
        """
        Deletes an existing collection.

        :param str collection_id: The id of the collection to be deleted.

        :returns: None

        - **Sample**

        .. code-block:: python

                client.delete_collection(collection_id)

        """
        if not collection_id:
            raise MissingMandatoryArgumentException('collection_id')

        collection = self._get_collection(collection_id=collection_id)
        collection.delete_collection(collection_id=collection_id)
            
    def get_packages_for_collection(self, collection_id, count=100):
        """
        Returns a list of all packages grouped under a collection.

        :param str collection_id: The id of the collection.
        :param int count: Optional. Count of packages to be returned. Defaults to 100.

        :returns: List of all packages grouped under the collection.
        :rtype: list[collections.namedtuple]

        - **Sample**

        .. code-block:: python

                collection_id = 'my-collection-id'
                packages = client.get_packages_for_collection(collection_id, count=100)
        """
        if not collection_id:
            raise MissingMandatoryArgumentException('collection_id')
            
        ensure_count_is_valid(count)

        collection = self._get_collection(collection_id=collection_id)

        packages = collection.collection_packages(page_size=count).get_entities(rel="package")
        return [siren_to_entity(p) for p in packages]
